// vincia06.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to run H+jet events at LHC 7 TeV
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (C) 2018 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia/Vincia.h"
using namespace Pythia8;
using namespace Vincia;

// ROOT: include files for linking to ROOT, uncomment as necessary
// #include "TROOT.h"
// #include "TDirectory.h"
// #include "TFile.h"
// #include "TH1.h"
// #include "TH2.h"

// Main Program
int main() {

  //************************************************************************
  // Define Pythia 8 generator with plug-in

  Pythia pythia;
  VinciaPlugin vincia(&pythia);
  // Switch FSR outside resonance decays back to Pythia
  pythia.setShowerPtr(&vincia.fsrShower,0,0);    

  // Read user settings from file
  pythia.readFile("vincia06.cmnd");

  //************************************************************************

  // Shorthands
  Event& event = pythia.event;
  Settings& settings = pythia.settings;

  // Extract settings to be used in the main program.  
  int    nEvent    = settings.mode("Main:numberOfEvents");
  bool   vinciaOn  = settings.flag("Vincia");
  int    verbose   = settings.mode("Vincia:verbose");

  //************************************************************************

  // Initialize
  vincia.init();

  //************************************************************************

  // Define PYTHIA histograms
  // Rootlink: to use ROOT instead, HIST -> TH1F ("tag",...)

  // Number of quarks, partons, and charged particles
  double nQsum=0.0;
  double nPsum=0.0;
  double nCsum=0.0;
  Hist HnQuarks("nQuarks", 50, -0.5, 49.5);
  Hist HnPartons("nPartons", 100, -0.5, 199.5);
  Hist HnCharged("nCharged", 100, -0.5, 399.5);

  //************************************************************************

  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage

  // (Optionally) re-initialize random number engine before event loop
  const int initseed = settings.mode("Random:seed");
  vincia.rndmPtr->init(initseed);
  
  // Counter for negative-weight events
  double evwt=1.0;

  // Begin event loop
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

    // Generate event
    if (!pythia.next()) continue;

    // Check for weights
    if (vinciaOn) evwt = vincia.weight();

    // Count FS charged hadrons, partons, and quarks
    int nc=0;
    int np=0;
    int nq=0;
    for (int i=1;i<event.size();i++) {
      // Count up final-state charged hadrons
      if (event[i].isCharged() && event[i].status() > 80) nc++;
      // Find last parton-level partons
      int iDau1 = event[i].daughter1();
      if (iDau1 == 0 || abs(event[iDau1].status()) > 80) {
	// Count up partons and quarks
	if (event[i].isQuark() || event[i].isGluon() ) np++;
	if (event[i].isQuark()) nq++;	
        
      }
    }
    HnQuarks.fill(0.5*nq,evwt);
    HnPartons.fill(1.0*np,evwt);
    HnCharged.fill(1.0*nc,evwt);
    nQsum += 0.5*nq;
    nPsum += np;
    nCsum += nc;    

  }

  //************************************************************************

  // POST-RUN FINALIZATION 
  // Normalization, Statistics, Output

  //Normalize histograms to effective number of positive-weight events.
  double normFac=1.0/(nEvent);
  HnQuarks    *= normFac;
  HnPartons   *= normFac;
  HnCharged   *= normFac;
    
  // Print a few histograms
  cout<<HnQuarks<<endl;
  cout<<" ^nQuarks"<<endl;
  cout<<HnPartons<<endl;
  cout<<" ^nPartons"<<endl;
  cout<<HnCharged<<endl;
  cout<<" ^nCharged"<<endl;
  
  cout<<endl;
  cout<<fixed;
  cout<< " <nGluonSplit> = "<<num2str(nQsum * normFac-1.0)<<endl;
  cout<< " <nPartons>    = "<<num2str(nPsum * normFac)<<endl;
  cout<< " <nCharged>    = "<<num2str(nCsum * normFac)<<endl;  
  cout<<endl;

  // Print out Vincia internal histograms
  if (vinciaOn && verbose >= 3) vincia.printHistos();

  // Print out Vincia end-of-run information
  pythia.stat();
  if (vinciaOn) vincia.printInfo();

  //Output tables of histograms to file?   
  /*
  HnCharged.table(mycout);
  */

  // Done.
  return 0;
}
