// vincia10-root.cc is a part of the VINCIA plugin to PYTHIA 8
// This is a simple test program to study energy scaling of ee observables
// Uses VINCIA for showering and matching and PYTHIA 8 for all else
// Copyright (C) 2018 Peter Skands

// PYTHIA and VINCIA include files
#include "Vincia/Vincia.h"
// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot/VinciaRoot.h"
#include "VinciaRoot/PythiaRoot.h"

// Use Pythia8 and Vincia namespaces, for convenience.
using namespace Pythia8; 
using namespace Vincia; 

static const int nEnergies = 15;
static const double eCM[nEnergies] = {14.,22.,29.,34.8,43.6,57.8,
                                      91.2,91.2,91.2,91.2,
                                      133.,161.,183.,189.0,200.};

static const int nEnergiesKch = 10;
static const double eKch[nEnergiesKch] = {14.,22.,34.8,43.6,91.2,91.2,
                                          133.,161.,183.,189.};

static const int nEnergiesLam = 8;
static const double eLam[nEnergiesLam] = {14.,34.8,91.2,91.2,91.2,91.2,
                                          133.,189.}; 

// Main Program
int main() {

  //************************************************************************
  // PYTHIA and VINCIA objects
  //************************************************************************

  // Create Pythia and Vincia objects. 
  Pythia pythia; 
  VinciaPlugin vincia(&pythia);

  // Read user settings from file
  pythia.readFile("vincia10.cmnd");

  //************************************************************************
  // Define some useful pointers and shorthands for later on
  //************************************************************************

  // Shorthand for event record
  Event* event = &pythia.event;

  // Some useful pointers
  Settings* settingsPtr   = vincia.settingsPtr;        // to settings database  

  // Extract settings to be used in the main program.  
  int    nEventSav        = settingsPtr->mode("Main:numberOfEvents");  

  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");
  int    verbose          = settingsPtr->mode("Vincia:verbose");

  // Matching settings
  int    matchingLO       = settingsPtr->mode("Vincia:matchingLO");
  bool   matching         = matchingLO >= 1;

  // hadronization settings
  bool   hadronLevel      = settingsPtr->flag("HadronLevel:all");

  // Data path directory, extract from VINCIADATA
  string VINCIA_SHARE     = vincia.VINCIA_SHARE;
  string VINCIA_VAL       = VINCIA_SHARE+"validation/";
  string VINCIA_DAT       = VINCIA_SHARE+"data/";

  //************************************************************************
  // VinciaRoot
  // Create VinciaRoot object and set plot labels, legends, etc
  // Also define names of files to save and read histograms to/from
  //************************************************************************

  // Create VinciaRoot object and set default behavior to logarithmic y axes
  VinciaRoot vroot(settingsPtr); 
  vroot.init();
  
  // Set default to linear Y axis for all plots
  vroot.setLogY("all",false);

  // Tell VinciaRoot how verbose you want it to be
  vroot.setVerbose("all",verbose);
  vroot.setShowLegendChi2("all",false);
  
  // Optionally set left/right white-on-black labels
  vroot.setLeftLabel("all","ee#rightarrowhadrons");

  // Set first update and save event
  double iSave   = 256;  

  // Set MC names and versions
  string pyVer = num2str(settingsPtr->parm("Pythia:versionNumber"),5);  
  string vcVer = num2str(settingsPtr->parm("Vincia:versionNumber"),5);  
  string mcVersions = "Vincia " + vcVer 
    + (matching ? " + MadGraph 4.4.26" : "") + " + Pythia "+pyVer;

  // Set label for this run
  string mcLabel = (vinciaOn) ? "Vincia" : "Pythia";
  if (!hadronLevel) mcLabel += " partons";
  // Set VinciaRoot plot labels, legends, etc
  vroot.setLegend("all",mcLabel);
  
  // Compare current run to files with the following prefix(es)
  vector<string> mcFileRead;
  mcFileRead.push_back(string(VINCIA_VAL+"vincia2/vincia10").c_str());
  mcFileRead.push_back(string(VINCIA_VAL+"pythia/vincia10").c_str());
  //mcFileRead.push_back(string(VINCIA_VAL+"vincia1/vincia10").c_str());

  // Set VinciaRoot plot labels, legends, etc
  vroot.setLegend("all",mcLabel);

  //************************************************************************
  // Initialize VINCIA (automatically initializes PYTHIA) 
  //************************************************************************

  // Initialize
  vincia.init();

  //************************************************************************
  // Book Histograms
  //************************************************************************

  // Set default plot sizes
  double plotSizeX = 420;
  double plotSizeY = 420;
  vroot.setPlotSize(plotSizeX,plotSizeY);

  // Histogram bin and range variables
  int nBins;
  double xLo, xHi;

  // INEL
  nBins = nEnergies;
  xLo   = -0.5;
  xHi   = xLo + nEnergies;
  vector<TH1D*> VNch;
  VNch.push_back(new TH1D("Nch","Average Charged Multiplicity vs ECM",
                          nBins, xLo, xHi));
  VNch[0]->GetYaxis()->SetTitle("<n_{Ch}>");
  VNch[0]->GetXaxis()->SetTitle("E_{cm} (not to scale)");
  for (int iBin=1; iBin<=nBins; ++iBin) 
    VNch[0]->GetXaxis()->SetBinLabel(iBin,string(num2str(int(eCM[iBin-1]+0.49))).c_str());
  VNch[0]->SetMaximum(65.01); 
  VNch[0]->SetMinimum(0.0); 
  VNch[0]->GetXaxis()->SetLabelOffset(0.01);
  VNch[0]->GetXaxis()->SetLabelSize(0.05);
  VNch[0]->GetXaxis()->SetTitleOffset(1.6);
  vroot.createDisplay("Nch",VNch);
  vroot.setNormalize("Nch",false);
  vroot.setDataFile("Nch",VINCIA_DAT+"ee-All-Nch.dat",25,1,40,50,60,70);  
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("Nch",mcFileRead[iRead]+"-Nch.dat");

  // 
  nBins = nEnergiesKch;
  xLo   = -0.5;
  xHi   = xLo + nBins;
  vector<TH1D*> VnKch;
  VnKch.push_back(new TH1D("NK","Average K^{+/-} Multiplicity vs ECM",
                          nBins, xLo, xHi));
  VnKch[0]->GetYaxis()->SetTitle("<n_{K}>");
  VnKch[0]->GetXaxis()->SetTitle("E_{cm} (not to scale)");
  for (int iBin=1; iBin<=nBins; ++iBin) 
    VnKch[0]->GetXaxis()->SetBinLabel(iBin,string(num2str(int(eKch[iBin-1]+0.49))).c_str());
  VnKch[0]->SetMaximum(8.01); 
  VnKch[0]->SetMinimum(0.0); 
  VnKch[0]->GetXaxis()->SetLabelOffset(0.01);
  VnKch[0]->GetXaxis()->SetLabelSize(0.05);
  VnKch[0]->GetXaxis()->SetTitleOffset(1.5);
  vroot.createDisplay("nKch",VnKch);
  vroot.setNormalize("nKch",false);
  vroot.setDataFile("nKch",VINCIA_DAT+"ee-All-nKch.dat",25,1,40,50,60,70);  
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("nKch",mcFileRead[iRead]+"-nKch.dat");
  
  // 
  nBins = nEnergiesLam;
  xLo   = -0.5;
  xHi   = xLo + nBins;
  vector<TH1D*> VnLam;
  VnLam.push_back(new TH1D("NKLam","Average #Lambda^{0} Multiplicity vs ECM",
                          nBins, xLo, xHi));
  VnLam[0]->GetYaxis()->SetTitle("<n_{#Lambda}>");
  VnLam[0]->GetXaxis()->SetTitle("E_{cm} (not to scale)");
  for (int iBin=1; iBin<=nBins; ++iBin) 
    VnLam[0]->GetXaxis()->SetBinLabel(iBin,string(num2str(int(eLam[iBin-1]+0.49))).c_str());
  VnLam[0]->SetMaximum(1.5); 
  VnLam[0]->SetMinimum(0.0); 
  VnLam[0]->GetXaxis()->SetLabelOffset(0.01);
  VnLam[0]->GetXaxis()->SetLabelSize(0.05);
  VnLam[0]->GetXaxis()->SetTitleOffset(1.5);
  vroot.createDisplay("nLam",VnLam);
  vroot.setNormalize("nLam",false);
  vroot.setDataFile("nLam",VINCIA_DAT+"ee-All-nLam.dat",25,1,40,50,60);  
  for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
    vroot.addHisFile("nLam",mcFileRead[iRead]+"-nLam.dat");
  

  //***************************************************************************
  // EVENT GENERATION LOOP
  // Generation, event-by-event printout, analysis, and storage
  //***************************************************************************

  // Loop over CM energies
  double eCMold = 0.0;
  for (int iECM=0; iECM<nEnergies; ++iECM) {
    double eCMnow = eCM[iECM];
    // Skip if several identical energies in succession
    if (eCMnow == eCMold) continue;
    eCMold = eCMnow;
    double nEvent = nEventSav;
    // Generate more statistics for lowest-energy points
    if (eCMnow <= 40.) nEvent *= 2.;
    // Generate more statistics for energies with Lambda
    bool hasLam    = false;
    for (int jECM=0; jECM<nEnergiesLam; ++jECM) 
      if (eLam[jECM] == eCMnow) hasLam = true;
    if (hasLam) nEvent *= 4.;

    // Initialize PYTHIA for this CM energy
    pythia.readString("Beams:eCM = "+num2str(eCMnow));
    vincia.init();
    
    // Begin event loop
    double nCsum       = 0.0;
    double nC2sum      = 0.0;
    double nKchSum     = 0.0;
    double nKch2sum    = 0.0;
    double nLamSum  = 0.0;
    double nLam2sum = 0.0;
    double iSaveNow = iSave;
    for (int iEvent = 0; iEvent < nEvent; ++iEvent) {

      // Generate next event
      if (!pythia.next()) continue;
      event = &pythia.event;
      
      // Count up average charged particles
      double nC      = 0 ;
      double nKch    = 0.;
      double nLam    = 0.;
      for (int i=1;i<event->size();i++) {
        int idAbs    = abs((*event)[i].id());
        if ((*event)[i].isCharged() && (*event)[i].status() > 80) ++nC; 
        if (idAbs == 321) nKch += 1;
        if (idAbs == 3122) nLam += 1;
      }
      nCsum       += nC;
      nC2sum      += pow2(nC);
      nKchSum     += nKch;
      nKch2sum    += pow2(nKch);
      nLamSum     += nLam;
      nLam2sum    += pow2(nLam);

      // Periodically update displays
      if ( iEvent == int(iSaveNow) || 
           (iEvent+251)%200 == 0 || iEvent == nEvent-1) {
        // Fill all bins that matches this CM energy
        for (int jECM=0; jECM<nEnergies; ++jECM) {
          if (eCM[jECM] != eCMnow) continue;
          double avg  = nCsum/(iEvent+1);
          double rms  = sqrt(nC2sum/iEvent+1);
          double sig2 = max(0.,pow2(rms)-pow2(avg));
          double err  = sqrt(sig2/(iEvent+1));
          VNch[0]->SetBinContent(jECM+1,avg);
          VNch[0]->SetBinError(jECM+1,err);
        }
        // Fill all Kch bins that matches this CM energy
        for (int jECM=0; jECM<nEnergiesKch; ++jECM) {
          if (eKch[jECM] != eCMnow) continue;
          double avg  = nKchSum/(iEvent+1);
          double rms  = sqrt(nKch2sum/iEvent+1);
          double sig2 = max(0.,pow2(rms)-pow2(avg));
          double err  = sqrt(sig2/(iEvent+1));
          VnKch[0]->SetBinContent(jECM+1,avg);
          VnKch[0]->SetBinError(jECM+1,err);
        }
        // Fill all Kch-from-Lambda bins that matches this CM energy
        for (int jECM=0; jECM<nEnergiesLam; ++jECM) {
          if (eLam[jECM] != eCMnow) continue;
          double avg  = nLamSum/(iEvent+1);
          double rms  = sqrt(nLam2sum/iEvent+1);
          double sig2 = max(0.,pow2(rms)-pow2(avg));
          double err  = sqrt(sig2/(iEvent+1));
          VnLam[0]->SetBinContent(jECM+1,avg);
          VnLam[0]->SetBinError(jECM+1,err);
        }
      }
      // Periodically update and save displays
      vroot.update(iEvent);       
      
    }
    // End of run
    // Save displays with base name mcFile and file type .eps
    vroot.update(2*nEvent);

    // Print out Vincia internal histograms
    if (vinciaOn && verbose >= 2) vincia.printHistos();
    // Print out Vincia end-of-run information
    pythia.stat();
    if (vinciaOn) vincia.printInfo();

  } // end loop over CM energies
  
  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;
  
  return 0;
}
