// vincia23-root.cc is a part of the VINCIA event generator.
// Copyright (C) 2018 Peter Skands.
// VINCIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This is a simple test program (adapted from the PYTHIA example main05.cc)
// It studies Z bosons at the LHC, summing over all contributing channels

// PYTHIA and VINCIA include files and namespaces
#include "Vincia/Vincia.h"
using namespace Pythia8;
using namespace Vincia;

// ROOT: include files for linking to ROOT, uncomment as necessary
#include "VinciaRoot/VinciaRoot.h"
#include "VinciaRoot/PythiaRoot.h"

// Global on/off switches for plots:
const bool OPTcreateDisplays = true;
const bool OPTshowWeightHistogram = true;

int main() {

  // Generator. LHC process and output selection. Initialization.
  Pythia pythia;
  VinciaPlugin vincia(&pythia);

  // Read user settings from file
  pythia.readFile("vincia23.cmnd");

  // Some useful pointers
  Settings* settingsPtr   = vincia.settingsPtr;        // to settings database  
  // Extract settings to be used in the main program.  
  double eCM              = settingsPtr->parm("Beams:eCM");
  int    nEvent           = settingsPtr->mode("Main:numberOfEvents");  
  
  // Vincia settings
  bool   vinciaOn         = settingsPtr->flag("Vincia");

  // MPI and hadronization settings
  bool   mpi              = settingsPtr->flag("PartonLevel:MPI");
  bool   hadronLevel      = settingsPtr->flag("HadronLevel:all");
  
  // Data path directory, extract from VINCIADATA
  string VINCIA_SHARE     = vincia.VINCIA_SHARE;
  string VINCIA_VAL       = VINCIA_SHARE+"validation/";
  string VINCIA_DAT       = VINCIA_SHARE+"data/";

  //************************************************************************
  // Initialize VINCIA (automatically initializes PYTHIA) 
  //************************************************************************

  vincia.init();

  //************************************************************************
  // VinciaRoot
  // Create VinciaRoot object and set plot labels, legends, etc
  // Also define names of files to save and read histograms to/from
  //************************************************************************

  // Create VinciaRoot object and set default behavior to logarithmic y axes
  VinciaRoot vroot(settingsPtr);
  vroot.init();
  vroot.setLogY();
  // Set range on ratio plots (default is [0.5,1.5])
  vroot.setRatioRange("all",0.5,1.5);
  vroot.setShowLegendChi2("all",false);
  // Do not show uncertainty breakdowns
  vroot.setShowRelUnc("all",false);
  
  // Optionally set left/right white-on-black labels
  vroot.setLeftLabel("all","pp#rightarrowZ#rightarrowleptons");
  string eCMstring = num2str(int(eCM)) + " GeV";
  vroot.setRightLabel("all",eCMstring);

  // Compare current run to files with the following prefix(es)
  vector<string> mcFileRead;
  mcFileRead.push_back(string(VINCIA_VAL+"vincia2/vincia23").c_str());
  mcFileRead.push_back(string(VINCIA_VAL+"pythia/vincia23").c_str());
  
  // Examples of how to set plot information and define read/save file names
  string mcLabel;
  // Example 1: PYTHIA runs
  if (!vinciaOn) {
    mcLabel    = "Pythia";
    if (settingsPtr->parm("SpaceShower:renormMultFac") > 1.2) {
      mcLabel += " #muHi";
    } else if (settingsPtr->parm("SpaceShower:renormMultFac") < 0.8) {
      mcLabel += " #muLo";
    }
  }
  // Example 2: VINCIA runs
  else {
    mcLabel    = "Vincia";
  }
  // Example 3: check if special settings results (hadronization or MPI OFF)
  if (pythia.parm("SpaceShower:renormMultFac") < 0.8) {
    mcLabel += " #mu=0.5p_{T}";
  }
  else if (pythia.parm("SpaceShower:renormMultFac") > 1.2) {
    mcLabel += " #mu=2.0p_{T}";
  }  
  if (!hadronLevel) {
    mcLabel += " noHad";
  }
  if (!mpi) {
    mcLabel += " noMpi";
  }

  // Set label for this run
  vroot.setLegend("all",mcLabel);
  
  //************************************************************************
  // Book Histograms
  //************************************************************************

  // Are we doing multiple event weights (e.g. uncertainty variations)
  int    nWeights         = vincia.nWeights();

  // Define a few simple PYTHIA Histograms.
  Hist HmZ("mZ", 100,50.,150.);
  Hist HmHat("mHat",100,0.,1000.);
  
  // Histogram of Event Weights
  double xLo = -1.1;
  double xHi =  7.1;
  double binWid = 0.02;
  int nBins = (xHi-xLo)/binWid;
  // Define main histogram; set axis labels and default display range
  vector<TH1D*> Vweights;
  Vweights.push_back(new TH1D("weights","Weights",nBins,xLo,xHi));
  Vweights[0]->GetXaxis()->SetTitle("weight");
  Vweights[0]->GetYaxis()->SetTitle("N");
  Vweights[0]->SetMinimum(1e-4);
  Vweights[0]->SetMaximum(100.0);
  // Bin weight = inverse bin size
  double wtWeights = 1./binWid;
  // Add more copies if doing multiple weights (e.g, uncertainty variations)
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    Vweights.push_back(new TH1D(string("weigths"+num2str(iWeight,0)).c_str(),
			       " ",nBins,xLo,xHi));
  // Create display
  if (OPTcreateDisplays && OPTshowWeightHistogram) {
    vroot.createDisplay("weights",Vweights,400,300);
    vroot.setNormalize("weights");
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) {
      vroot.addHisFile("weights",mcFileRead[iRead]+"-weights.dat");
    }
  }

  // Histogram of pTZ tail spectrum
  nBins  = 396;
  xLo    = 8.0;
  xHi    = 800.0;
  // Define main histogram; set axis labels and default display range
  vector<TH1D*> VpTZtail;
  VpTZtail.push_back(new TH1D("pTZtail","p_{TZ} (Born)",nBins,xLo,xHi));
  VpTZtail[0]->GetXaxis()->SetTitle("p_{TZ} [GeV]");
  VpTZtail[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dp_{T}");
  VpTZtail[0]->SetMinimum(0.0000001);
  VpTZtail[0]->SetMaximum(10.);
  // Bin weight = inverse bin size
  double wpTZtail = nBins/(xHi-xLo);
  // Add more copies if doing multiple weights (e.g, uncertainty variations)
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    VpTZtail.push_back(new TH1D(string("pTZtail"+num2str(iWeight,0)).c_str(),
			       " ",nBins,xLo,xHi));  
  
  // Create second uncertainty-variation vector
  vector<TH1D*> VpTZtail2;
  VpTZtail2.push_back(new TH1D("pTZtailB","p_{TZ} (Born)",nBins,xLo,xHi));
  // Uncertainty variations
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    VpTZtail2.push_back(new TH1D(string("pTZtailB"+num2str(iWeight,0)).c_str(),
			       " ",nBins,xLo,xHi));
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("pTZtail",VpTZtail,400,500);
    //vroot.addTH1DPtr("pTZtail",VpTZtail2);
    vroot.setDataFile("pTZtail",string(VINCIA_DAT+"Atlas-7000-pTZ.dat").c_str(),10,20,30,40,0,552,652,752);
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) {
      vroot.addHisFile("pTZtail",mcFileRead[iRead]+"-pTZtail.dat");
    }
    vroot.setRebinLikeData("pTZtail");
    vroot.setLogX("pTZtail");
    vroot.setNormalize("pTZtail",true);
  } 
 
  // Histogram of Spectrum after the first recoil
  vector<TH1D*> VpTZtailOne;
  VpTZtailOne.push_back(new TH1D("pTZtailOne","p_{TZ} (Born, after 1st Branching)",nBins,xLo,xHi));
  double wpTZtailOne = nBins/(xHi-xLo);
  VpTZtailOne[0]->GetXaxis()->SetTitle("p_{TZ} [GeV]");
  VpTZtailOne[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dp_{T}");
  VpTZtailOne[0]->SetMinimum(0.0000001);
  VpTZtailOne[0]->SetMaximum(10.);
  // Uncertainty variations
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    VpTZtailOne.push_back(new TH1D(string("pTZtailOne"+num2str(iWeight,0)).c_str(),
			       " ",nBins,xLo,xHi));
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("pTZtailOne",VpTZtailOne,400,500);
    vroot.setDataFile("pTZtailOne",string(VINCIA_DAT+"Atlas-7000-pTZ.dat").c_str(),10,20,30,40,0,552,652,752);
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) 
      vroot.addHisFile("pTZtailOne",mcFileRead[iRead]+"-pTZtailOne.dat");
    vroot.setRebinLikeData("pTZtailOne");
    vroot.setLogX("pTZtailOne");
    vroot.setNormalize("pTZtailOne",true);
  }
  
  // Histogram of pTZ Peak spectrum
  nBins = 63.0;
  xLo =   0.0;
  xHi =  42.0;
  vector<TH1D*> VpTZpeak;
  VpTZpeak.push_back(new TH1D("pTZpeak","p_{TZ} (Born)",nBins,xLo,xHi));
  VpTZpeak[0]->GetXaxis()->SetTitle("p_{TZ} [GeV]");
  VpTZpeak[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dp_{T}");
  VpTZpeak[0]->SetMinimum(-0.02);
  VpTZpeak[0]->SetMaximum(0.095);
  double wpTZpeak = nBins/(xHi-xLo);
  // Uncertainty variations
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    VpTZpeak.push_back(new TH1D(string("pTZpeak"+num2str(iWeight,0)).c_str(),
			       " ",nBins,xLo,xHi));
  // Second Uncertainty variations
  vector<TH1D*> VpTZpeak2;
  VpTZpeak2.push_back(new TH1D("pTZpeakB","p_{TZ} (Born)",nBins,xLo,xHi));
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    VpTZpeak2.push_back(new TH1D(string("pTZpeakB"+num2str(iWeight,0)).c_str(),
			       " ",nBins,xLo,xHi));
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("pTZpeak",VpTZpeak,400,500);
    //vroot.addTH1DPtr("pTZpeak",VpTZpeak2);
    vroot.setDataFile("pTZpeak",string(VINCIA_DAT+"Atlas-7000-pTZ.dat").c_str(),10,20,30,40,0,552,652,752);
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) {
      vroot.addHisFile("pTZpeak",mcFileRead[iRead]+"-pTZpeak.dat");
    }
    vroot.setRebinLikeData("pTZpeak");
    vroot.setNormalize("pTZpeak",true);
  }
  
  // Histogram of spectrum after the first recoil
  vector<TH1D*> VpTZpeakOne;
  VpTZpeakOne.push_back(new TH1D("pTZpeakOne","p_{TZ} (Born, after 1st Branching)",nBins,xLo,xHi));
  VpTZpeakOne[0]->GetXaxis()->SetTitle("p_{TZ} [GeV]");
  VpTZpeakOne[0]->GetYaxis()->SetTitle("1/#sigma d#sigma/dp_{T}");
  VpTZpeakOne[0]->SetMinimum(-0.02);
  VpTZpeakOne[0]->SetMaximum(0.095);
  double wpTZpeakOne = nBins/(xHi-xLo);
  // Uncertainty variations
  for (int iWeight=1; iWeight<nWeights; iWeight++) 
    VpTZpeakOne.push_back(new TH1D(string("pTZpeakOne"+num2str(iWeight,0)).c_str(),
			       " ",nBins,xLo,xHi));
  // Create display and compare to ATLAS data
  if (OPTcreateDisplays) {
    vroot.createDisplay("pTZpeakOne",VpTZpeakOne,400,500);
    vroot.setDataFile("pTZpeakOne",string(VINCIA_DAT+"Atlas-7000-pTZ.dat").c_str(),10,20,30,40,0,552,652,752);
    // Read in any saved runs
    for (int iRead=0; iRead<int(mcFileRead.size()); ++iRead) {
      vroot.addHisFile("pTZpeakOne",mcFileRead[iRead]+"-pTZpeakOne.dat");
    }
    vroot.setNormalize("pTZpeakOne",true);
  }

  //************************************************************************
  // Event Loop
  //************************************************************************
  
  // Begin event loop. Generate event. Skip if error.
  double wSum  = 0.0;
  double wSumA = 0.0;
  double wSumB = 0.0;
  double wSumLo  = 0.0;
  double wSumLoA = 0.0;
  double wSumLoB = 0.0;
  for (int iEvent = 0; iEvent < nEvent; ++iEvent) {
    if (!pythia.next()) continue;

    // Reference to event
    Event& event=pythia.event;

    double wt = vincia.weight();
    HmHat.fill(pythia.info.mHat(),wt);
    HmZ.fill(pythia.process[5].m(),wt);
    for (int iWeight=0; iWeight<nWeights; ++iWeight) {
      Vweights[iWeight]->Fill(vincia.weight(iWeight),wtWeights);
      //cout<<" w"<<iWeight<<" = "<<vincia.weight(iWeight)<<"   ";
    }
    //cout<<endl;
    
    // Find the first and last Z boson
    int iZ = 1;
    int iZ1 = 0;
    for (int i=5; i<pythia.event.size(); ++i)
      if (pythia.event[i].id() == 23) {
	iZ = i; 
	if (iZ1 == 0 && pythia.event[i].pT() > 0.001
	    && abs(pythia.event[i].status()) != 62) iZ1 = i;
      }
    
    // Check daughter lepton momenta
    int ida1 = event[iZ].daughter1();
    int ida2 = event[iZ].daughter2();
    if (abs(event[ida1].eta()) > 2.4 || event[ida1].pT() < 20.) continue;
    if (abs(event[ida2].eta()) > 2.4 || event[ida2].pT() < 20.) continue;    
    // Plot pTZ
    double pTZborn = pythia.event[iZ].pT();
    for (int iWeight=0; iWeight<nWeights; ++iWeight) {
      if (iWeight <= 2)
	VpTZpeak[iWeight]->Fill(pTZborn,wpTZpeak*vincia.weight(iWeight));
      if (iWeight == 0 || iWeight >= 3)
	VpTZpeak2[iWeight]->Fill(pTZborn,wpTZpeak*vincia.weight(iWeight));
      if (iWeight <= 2)
	VpTZtail[iWeight]->Fill(pTZborn,wpTZtail*vincia.weight(iWeight));      
      if (iWeight == 0 || iWeight >= 3)
	VpTZtail2[iWeight]->Fill(pTZborn,wpTZtail*vincia.weight(iWeight));    
    }
    
    wSum += vincia.weight(0);
    if (nWeights >= 2) wSumA += vincia.weight(1);
    if (nWeights >= 3) wSumB += vincia.weight(2);
    double pTorg = pythia.event[iZ].pT();
    if (pTorg < 5.0) {
      wSumLo  += vincia.weight(0);
      if (nWeights >= 2) wSumLoA += vincia.weight(1);
      if (nWeights >= 3) wSumLoB += vincia.weight(2);
    }

    // Plot pTZ after first emission
    double pTZ1 = pythia.event[iZ1].pT();
    if (iZ1 == 0) pTZ1 = 0.001;
    for (int iWeight=0; iWeight<nWeights; ++iWeight) {
      VpTZpeakOne[iWeight]->Fill(pTZ1,wpTZpeakOne*vincia.weight(iWeight));
      VpTZtailOne[iWeight]->Fill(pTZ1,wpTZtailOne*vincia.weight(iWeight));
    }

    // Periodically save displays
    vroot.update();
    /*
    if ( iEvent >= iSave ) {
      iSave       = vroot.nextSave();
      cout<<"Next save at event "<<int(iSave + 0.5)<<endl;
      if (nWeights >= 0) {
	cout<<"<w>     = "<<wSum/(iEvent+1)<<" "<<wSumA/(iEvent+1)<<" "<<wSumB/(iEvent+1)<<endl;
	cout<<"P(pT<5) = "<<wSumLo/wSum<<" "<<wSumLoA/wSumA<<" "<<wSumLoB/wSumB<<endl;
	cout<<"R(pT<5) = "<<wSumLoA/wSumA/(wSumLo/wSum)<<" "<<wSumLoB/wSumB/(wSumLo/wSum)<<endl;
      }
    }
    */
  }  
  
  // End-of-run update and save
  vroot.update(vroot.nextSave());       

  pythia.stat();
  cout<<HmHat<<HmZ;

  cout << endl << "Enter q to terminate (will kill displays)..." << endl;
  char tmp;cin>>tmp;

  // Done.
  return 0;
}
