// VinciaWeights.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2018 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Variations class.
  
#ifndef Vincia_VinciaWeights_H
#define Vincia_VinciaWeights_H

#include "Pythia8/Pythia.h"
#include "Vincia/VinciaCommon.h"
#include "Vincia/Branchelementals.h"

namespace Vincia {

  class VinciaWeights {

  public:

    // Constructor.
    VinciaWeights() {}

    // Destructor.
    ~VinciaWeights() {}

    // Initilize Pointers.
    bool initPtr(Settings* settingsPtrIn, Info* infoPtrIn,
      VinciaCommon* vinComPtrIn) {
      settingsPtr = settingsPtrIn;
      infoPtr     = infoPtrIn;
      vinComPtr   = vinComPtrIn;
      isInitPtr   = true;
      return true;
    }

    // Initilization.
    void init();

    // Check if particular weight exists.
    bool existsWeight(int iWeightIn=0) {
      if (iWeightIn >= 0 && iWeightIn < nWeightsSav) return true;
      else return false;
    }

    // Get weight of current event weight
    // Normally 1 for set 0 (user settings) 
    // Use iWeightIn > 0 to access uncertainty weights
    double weight(int iWeightIn=0) {
      if (existsWeight(iWeightIn)) return weightsSav[iWeightIn];
      return 0.0;
    }

    // Access the weight labels
    string weightLabel(int iWeightIn=0) {
      if (iWeightIn == 0) return "no variation";
      if(existsWeight(iWeightIn) && iWeightIn-1 < (int)varLabels.size())
        return varLabels[iWeightIn-1];
      return "";
    }

    // Functions to access by VINCIA plugin
    int  nWeights()         { return nWeightsSav; }
    bool reweightingOccurred() { return didReweight; }

    // Reset the weights
    void resetWeights(int nAccepted);

    // Scale all event weights
    void scaleWeightAll(double scaleFacIn) {
      for (int iWeight=0; iWeight<nWeightsSav; iWeight++)
        weightsSav[iWeight] *= scaleFacIn;
    }

    // Scale a particular event weight
    void scaleWeight(double scaleFacIn, int iWeightIn=0) {
      if (existsWeight(iWeightIn)) weightsSav[iWeightIn] *= scaleFacIn;
    }

    // Scale the uncertainty band weights
    void scaleWeightVar(vector<double> Paccept, bool accept, bool isHard,
      string shower = "", int type = -1);

    // Scale the uncertainty band weights if branching is accepted
    void scaleWeightVarAccept(vector<double> Paccept, string shower = "",
      int type = -1);

    // Scale the uncertainty band weights if branching is rejected
    void scaleWeightVarReject(vector<double> Paccept, string shower = "",
      int type = -1);

    // Biased kernels: reweight if branching is accepted
    void scaleWeightBiasAccept(double biasFac=1.) {
      if (biasFac == 1.0) return;
      else scaleWeightAll(1./biasFac);
    }

    // Biased kernels: reweight if branching is rejected
    void scaleWeightBiasReject(double PacceptUnbiased, double biasFac=1.) {
      if (biasFac == 1.0) return;
      if (biasFac > 1.0) {
	double rRej = (1. - PacceptUnbiased/biasFac)/(1. - PacceptUnbiased);
	scaleWeightAll(rRej);
      } else {
	double rRej = (1. - PacceptUnbiased)/(1. - biasFac*PacceptUnbiased);
	scaleWeightAll(rRej);
      }
    }
    
    // Main function to perform the reweighting due to MC overestimate
    // violations
    pair<double,double> doMCreweighting(vector<double>& Paccept);

    // Helper function for keyword evaluation
    int doVarNow(string keyIn, int iAntPhys, string type) ;

    // Helper function for antenna function
    double ant(double antIn, double cNSIn) {
      if (allowNegAnt) return (antIn+cNSIn);
      return max(TINYANT,antIn+cNSIn);
    }

    // Main function to perform the ISR uncertainty variations
    void doVariationsISR(vector<double>& Paccept, bool isHard, bool doMEC,
      BranchelementalISR* isrTrial, int iTrial, AlphaStrong* asPtr,
      double asUsr, double mu2usr, double mu2min, double mu2freeze,
      double asMax, bool isPolarized, double antUsr, double helSumUsr,
      double aHelUsr);

    // Main function to perform the FSR uncertainty variations
    void doVariationsFSR(vector<double>& Paccept, bool isHard, bool doMEC,
      BranchelementalFSR* fsrTrial, int iTrial, AlphaStrong* asPtr,
      double asUsr, double mu2usr, double mu2min, double mu2freeze,
      double asMax, bool isPolarized, double antUsr, double helSumUsr,
      double aHelUsr);

    // Main function to perform the weighting
    void doWeighting();

    // Verbose
    int verbose;

    // Pointers
    Settings*     settingsPtr;
    Info*         infoPtr;
    VinciaCommon* vinComPtr;

    // Internal flag
    bool isInitPtr;

    // Constants
    static const double TINYANT, PACCEPTVARMAX, MINVARWEIGHT;

    // Parameters taken from settings
    bool   allowReweighting, uncertaintyBands;
    vector<string> varLabels;
    vector< vector<string> > varKeys;
    vector< vector<double> > varVals;
    double deltaASmax, cNSpTmin;
    bool   varInMPIshowers, muSoftCorr, allowNegAnt;
    bool   printPaccVarViol, printReWeightViol;
 
    // Helper parameters
    vector<string> allKeywords;
    map<int,string> iAntToKeyFSR, iAntToKeyISR;
    double nWeightsSav;
    int    nReportWeight, nReportedWeight;
    bool   doAlphaSvar, doFiniteVar;

    // Weights (from input events, biasing, MC violations, or variations)
    vector<double> weightsSav; // Weights
    vector<double> weightsOld; // Weights from last step
    vector<double> weightsMax, weightsMin;
    // WeightSums
    vector<double> weightSum, weightSum2;
    // Contribution of current event
    vector<double> contribSum, contribSum2;

    // Counters:
    // - total number of weights = nr of events
    int nTotWeights;
    // - for initial and main weight (MC violation only)
    int nNonunityWeight, nNegativeWeight,
        nNonunityInitialWeight, nNegativeInitialWeight;
    // - for current event
    int nNonunityWeightNow, nNegativeWeightNow,
        nNonunityInitialWeightNow, nNegativeInitialWeightNow;

    // Flag to tell if reweighting due to MC violations occured
    bool didReweight;

    // doWeighting called for the first time?
    bool firstCall;

  };

}

#endif
