// MG5matrix.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2018 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Functionality to interface to FORTRAN matrix elements generated with MadGraph 5

#ifndef VINCIA_MG5matrix_H
#define VINCIA_MG5matrix_H

// Include Pythia and headers
#include "Pythia8/Basics.h"
#include "Pythia8/PythiaComplex.h"
#include "Pythia8/Event.h"
#include "Pythia8/Info.h"
#include "Pythia8/ParticleData.h"
#include "Pythia8/PythiaStdlib.h"
#include "Pythia8/Settings.h"
#include "Pythia8/StandardModel.h"
//Used for ostream_iterator, which is used to convert a vector of integers
//into a vector of doubles
#include <iostream>
#include <iterator>

// Include Vincia Standard Library
#include "Vincia/VinciaCommon.h"

namespace Vincia {

  const static bool DEBUG=false;
  
//==========================================================================

// The MG5matrix class contains the base class implementation for wrapping
// FORTRAN matrix elements generated by the 'output matrix' command of MG5

class MG5matrix {
    
 public:

  // Constructor
  MG5matrix() { initHelMap(); initColMap(); };
  
  // Destructor
  virtual ~MG5matrix() {};

  // Initialise (base class implementation)
  virtual void init() {};
  virtual void initHelMap() {};
  virtual void initColMap() {};
  
  // Define name of process in MadGraph format
  // Base class example is for dummy process a + bbar -> c + dbar + g
  virtual string name()  {return "abx_cdxg";};

  // Process list = vector of all process labels that this ME can handle
  virtual vector< vector<int> > processList() {
    vector<int> label;
    label.push_back(0);
    label.push_back(0);
    label.push_back(0);
    label.push_back(0);
    vector< vector<int> > labels;
    labels.push_back(label);
    return labels;
  }
  
  // Return number of legs
  virtual int nLegs() {return 0;}
  
  // Reset for new configuration
  // NOTE: should not normally need to be redefined in derived classes.
  virtual void reset() {
    icolSav  = -1;
    ihelSav  = -1;
    iVincia.clear();
  }
  
  // Translate from Vincia colour-ordering to MG 0 1 -> 2 3 ... N ordering.
  // iVincia takes an index in MadGraph's enumeration and returns the index in
  // Vincia's enumeration. Thus, eg iVincia[0] and iVincia[1] must return the
  // positions of the two initial-state particles in Vincia's list.
  // (Can return false if translation impossible / inconsistent input)
  // Inputs = idIn  : vector of ID codes orderd in Vincia colour-ordering
  //          int iA: index of first initial-state parton
  //          int iB: index of second initial-state parton
  virtual bool setPerm(vector<Particle> pIn, int iA, int iB=-1) {

    // Sanity check : pIn has the size expected from nLegs().
    if ((int)pIn.size() != nLegs()) cout<<name()
      <<"::setPerm(): Warning! size of input vector "
      <<pIn.size()<<" != "<<nLegs()<<" = nLegs()"<<endl;

    // Map incoming particles
    if (iB < 0) iVincia[0] = iA;
    else {
      // Order incoming particles by PDG code
      bool swap = false;
      if (pIn[iB].id() < pIn[iA].id()) swap = true;
      iVincia[0] = swap? iB : iA;
      iVincia[1] = swap? iA : iB;
    }
    // Create list of outgoing particles 
    vector<int> iFS;
    for (int i=0; i<(int)pIn.size(); ++i) {
      if (i == iA || i == iB) continue;
      iFS.push_back(i);
    }
    // Order final-state particles by PDG code
    int iMG=1;
    while (iFS.size() >= 2) {
      int iMin  = 0;
      int idMin = pIn[iFS[0]].id();
      for (int i=1; i<(int)iFS.size(); ++i) {
	int id = pIn[iFS[i]].id();
	if (id < idMin) {
	  iMin  = i;
	  idMin = id;
	}
      }
      iVincia[++iMG] = iFS[iMin];
      // Erase iMin and repeat
      iFS.erase(iFS.begin()+iMin);
    }
    // iFS size is now down to 1. Add last remaining element.
    iVincia[++iMG] = iFS[0];
    
    // VERBOSE DEBUG
    if (DEBUG) {
      // print map
      cout<<"  DEBUG "<<name()<<"::setPerm(): iVincia[iMG] ";
      for (int i=0; i<nLegs(); ++i) cout<<"["<<i<<"]="<<iVincia[i]<<" ";
      cout<<endl;
    }
    
    // Check which partons are colour triplets (so they can be ordered
    // appropriately in colour traces; MG puts qqbar at the back)
    for (iMG=0; iMG<nLegs(); ++iMG) {
      // Initial-state partons are crossed
      int sign = 1;
      if (iVincia[iMG] == iA || iVincia[iMG] == iB) sign = -1;
      // Check if this parton is a (crossed) triplet.
      // Note: PYTHIA's Particle::colType() returns zero unless the
      // corresponding particle's colType has been explicitly initialised.
      // Here, just implement simple check based on ID code
      int id = pIn[iVincia[iMG]].id();
      // (Here, should really check if particle is octet)
      if (id == 21 || id == 1000021) {
	isOct[iMG]  = true;
	isTrip[iMG] = false;
      // (Here, should really check if particle is antitriplet)
      } else if (id*sign < 0) {
	isTrip[iMG] = false;
	isOct[iMG]  = false;
      } else {
	isTrip[iMG] = true;
	isOct[iMG]  = false;
      }
      // VERBOSE DEBUG
      if (DEBUG) {
	cout<<"  DEBUG isTrip["<<iMG<<"] = "<<isTrip[iMG]
	    <<" isOct = "<<isOct[iMG]<<endl;
      }
    }
    
    // Done
    return true;
  }
  
  // Store momenta in MG5 format (from input in Pythia Particle format)
  // (Must be copied in derived classes so that pSav will be the
  // variable in the derived class, rather than the one in this base class)
  virtual void setMom(vector<Particle> pIn, bool permute=true) {
    double esum(0.);
    for (int i=0; i<nLegs(); ++i) {      
      if (permute) {
	pSav[i][0] = pIn[iVincia[i]].e();
	pSav[i][1] = pIn[iVincia[i]].px();
	pSav[i][2] = pIn[iVincia[i]].py();
	pSav[i][3] = pIn[iVincia[i]].pz();
      } else {
	pSav[i][0] = pIn[i].e();
	pSav[i][1] = pIn[i].px();
	pSav[i][2] = pIn[i].py();
	pSav[i][3] = pIn[i].pz();
      }
      if (i >= 2) esum += pSav[i][0];
    }
    // Check 2->2, then decay energy conservation
    if ( abs(esum - pSav[0][0] - pSav[1][0])/esum > 1.e-6)
      if ( abs(esum + pSav[1][0] - pSav[0][0])/esum > 1.e-6)
	cout<<name()<<"::setMom(): Warning! Input configuration does not appear to conserve energy"<<endl;
  }

  // Store momenta in MG5 format (from input in Pythia Vec4 format)
  // (Must be copied in derived classes so that pSav will be the
  // variable in the derived class, rather than the one in this base class)
  virtual void setMom(vector<Vec4> pIn, bool permute=true) {
    double esum(0.);
    for (int i=0; i<nLegs(); ++i) {      
      if (permute) {
	pSav[i][0] = pIn[iVincia[i]].e();
	pSav[i][1] = pIn[iVincia[i]].px();
	pSav[i][2] = pIn[iVincia[i]].py();
	pSav[i][3] = pIn[iVincia[i]].pz();
      } else {
	pSav[i][0] = pIn[i].e();
	pSav[i][1] = pIn[i].px();
	pSav[i][2] = pIn[i].py();
	pSav[i][3] = pIn[i].pz();
      }
    }
    // Check 2->2, then decay energy conservation
    if ( abs(esum - pSav[0][0] - pSav[1][0])/esum > 1e-3)
      if ( abs(esum + pSav[1][0] - pSav[0][0])/esum > 1e-3)
	cout<<name()<<"::setMom(): Warning! Input configuration does not appear to conserve energy"<<endl;
  }

  // Determine which IHEL to use
  // Input = hIn     : vector of helicities 
  //         permute : false if entries already ordered in MG5 enumeration
  //                   true if entries need to be mapped using permutation map
  virtual int setHel(vector<int> hVec, bool permute=true) {
    if (permute) {
      vector<int> hVecOrg(hVec);
      for (int i=0; i<(int)hVec.size(); ++i) hVec[i] = hVecOrg[iVincia[i]];
    }
    if (hMapSav.find(hVec) != hMapSav.end()) {
      ihelSav = hMapSav[hVec];
    } else {
      cout<<" MG5matrix: helicity configuration not found"<<endl;
      ihelSav = -1;
    }
    return ihelSav;
  }
  
  // Determine which ICOL to use
  // ccList contains list of colour contractions
  virtual int setIcol(vector< vector<int> > ccList, bool permute=true) {

    // VERBOSE DEBUG
    if (DEBUG) {
      cout<<"  DEBUG MG5matrix::setIcol(): input colour chain";
      for (int i=0; i<(int)ccList.size(); ++i) {
	cout<<" : ";
	for (int j=0; j<(int)ccList[i].size(); ++j)
	  cout<<ccList[i][j]<<" ";
      }
      cout<<endl;
    }
    
    // Convert each contraction to use MG enumeration, and update ccList.
    if (permute) {
      for (int iChain=0; iChain<(int)ccList.size(); ++iChain) {
	vector<int> ccVincia = ccList[iChain];
	vector<int> ccMG; ccMG.resize(ccVincia.size());
	for (int j=0; j<(int)ccVincia.size(); ++j) {
	  int iV = ccVincia[j];
	  // Find MG index corresponding to this iVincia (alternatively,
	  // could construct a reverse-lookup map in setPerm()) 
	  map<int,int>::iterator it;
	  for (it=iVincia.begin(); it!=iVincia.end(); ++it)
	    // Translate to MG's FORTRAN numbering from [1,N]
	    if (it->second == iV) {
	      ccMG[j] = it->first + 1;	      
	    }
	}
	// VERBOSE DEBUG
	if (DEBUG) {
	  cout<<" Trace begins at "<<ccMG[0]<<" isTrip = "
	      <<isTrip[ccMG[0]-1]<<" isOct = "<<isOct[ccMG[0]-1]<<endl;
	}
	// If this is a gluon ring, pull lowest MG index to front
	if (isOct[ccMG[0]-1]) {
	  int iMin   = 0;
	  int iMGmin = ccMG[0]; 
	  for (int i=1; i<(int)ccMG.size(); ++i) {
	    if (ccMG[i] < iMGmin) {
	      iMGmin = ccMG[i];
	      iMin   = i;
	    }
	  }
	  // Cyclically permute trace if needed
	  if (iMin != 0) {
	    ccMG.insert(ccMG.end(), ccMG.begin(), ccMG.begin()+iMin);
	    ccMG.erase(ccMG.begin(), ccMG.begin() + iMin);
	  }
	}
	// Madgraph puts q-qbar at end of traces
	else if (isTrip[ccMG[0]-1]) {
	  // Duplicate last element and put quark second-to-last
	  ccMG.push_back(ccMG[ccMG.size()-1]);
	  ccMG[ccMG.size()-2] = ccMG[0];
	  // Erase quark from front of vector
	  ccMG.erase(ccMG.begin());
	}
	// 
	ccList[iChain] = ccMG;
      }
    }

    // VERBOSE DEBUG
    if (DEBUG) {
      cout<<"  DEBUG MG5matrix::setIcol(): converted colour chain";
      for (int i=0; i<(int)ccList.size(); ++i) {
	cout<<" : ";
	for (int j=0; j<(int)ccList[i].size(); ++j)
	  cout<<ccList[i][j]<<" ";
      }
      cout<<endl;
    }

    // If several colour chains, concatenate, arranging by lowest index first
    // We won't be dealing with huge lists, so just a simple swap-and-repeat
    vector<int> colTrace;
    if (ccList.size() >= 2) {
      bool unordered = true;
      int iLoop = 0;
      while (unordered) {
	++iLoop;
	unordered = false;
	for (int i=1; i<(int) ccList.size(); ++i) {
	  if (ccList[i][0] < ccList[i-1][0]) {
	    unordered = true;
	    colTrace = ccList[i-1];
	    ccList[i-1] = ccList[i];
	    ccList[i]   = colTrace;
	  }
	}
	if (iLoop >= 50) {
	  cout<<" MG5matrix::setIcol(): problem stuck in loop."<<endl;
	  break;
	}
      }

      // VERBOSE DEBUG
      if (DEBUG) {
	cout<<"  DEBUG MG5matrix::setIcol(): reordered colour chain";
	for (int i=0; i<(int)ccList.size(); ++i) {
	  cout<<" : ";
	  for (int j=0; j<(int)ccList[i].size(); ++j)
	  cout<<ccList[i][j]<<" ";
	}
	cout<<endl;	
      }
    }
    
    colTrace = ccList[0];
    // Concatenate if needed.
    for (int i=1; i<(int)ccList.size(); ++i) 
      colTrace.insert(colTrace.end(), ccList[i].begin(), ccList[i].end());

    // Look to see if this colour contraction exists in colour map
    if (cMapSav.find(colTrace) != cMapSav.end())
      icolSav = cMapSav[colTrace];
    else {
      cout<<name()<<": ERROR! colour configuration not found"<<endl;
      icolSav = -1;
    }
    return icolSav;
  }
  
  // Return matrix element squared.
  // Base class returns error (-1).
  // Derived classes should wrap MG5 matrix element, see examples in .cc file.
  // Input : int mcMode = 0 (LC), 1 (VC), or 2 (FC)
  virtual double getME2(int) { return -1.; }

  // Print momenta (in MG enumeration)
  virtual void printMom() {    
    for (int i=0; i<nLegs(); ++i) {
      cout<<" pMG"<<i<<" = ";
      for (int j = 0; j<4; ++j) cout<<" "<<setw(10)<<pSav[i][j];
      cout<<endl;
    }
  }
  
 protected:
  
  // Map to return Vincia index from a given MG index, see setPerm();
  map<int,int> iVincia;
  
  // LC index: specifies which ICOL is to be used to define LC
  int icolSav;

  // Helicity-amplitude index: -1 = helicity sum
  int ihelSav;
  
  // MG5-ordered momenta: pSav[NLEGS][4]
  double pSav[5][4];

  // Map from helicity vectors to IHEL numbers
  map< vector<int>, int > hMapSav;

  // Map from MG indices to ICOL numbers
  // (Convention is that lowest index is pulled to front)
  map< vector<int>, int > cMapSav;
  
  // Map to tell whether partons (in MG numbering) are colour triplets
  // (Required since the default MG5 output puts qbar-q at the end of
  // T contractions)
  map< int, bool > isTrip, isOct;
  
};

//==========================================================================

} // end namespace Vincia

#endif

