// Antennae.cc is a part of the VINCIA plug-in to the PYTHIA 8
// generator.  VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2018 P. Skands.  Please respect the MCnet Guidelines
// for Event Generator Authors and Users,
// http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the
// AntennaSet and Antenna classes for both FSR and ISR, and
// some related global functions.

#include "Vincia/AntennaFunctions.h"

namespace Vincia {

//==========================================================================

// AntennaSetFSR class.
// This class is a simple container for a set of FF antenna functions

void AntennaSetFSR::init() {
  
  if (! isInitPtr) {
    printOut("AntennaSetFSR::init","Cannot initialize, pointers not set.");
    return;
  }
  
  // Set verbose level
  verbose = settingsPtr->mode("Vincia:verbose");
  
  // Create antenna objects
  antPtrs.resize(5);
  antPtrs[iQQemit]  = new QQEmitFF();
  antPtrs[iQGemit]  = new QGEmitFF();
  antPtrs[iGGemit]  = new GGEmitFF();
  antPtrs[iQGsplit] = new QGSplitFF();
  antPtrs[iGGsplit] = new GGSplitFF();
    
  // Loop through antPtrs and initialise
  for (vector<AntennaFSR*>::iterator antPtrIt = antPtrs.begin(); 
       antPtrIt != antPtrs.end(); ++antPtrIt) {
    
    // Initialize antenna pointers to PYTHIA 8 objects
    (*antPtrIt)->initPtr(functionsPtr, particleDataPtr, rndmPtr, settingsPtr);
    
    // Initialize antenna itself
    bool pass = (*antPtrIt)->init();
    
    // Optionally check antenna (singularities, positivity, etc)
    if (settingsPtr->flag("Vincia:checkAntennae")) 
      pass = pass && (*antPtrIt)->check();
    
    // Everything OK?
    if (pass) {
      if (verbose >= 2)
        printOut("Vincia::AntennaSetFSR::init","Added to antenna list: "
                 +(*antPtrIt)->humanName());
    } 
    else if (verbose >= 1) {
      printErr("Vincia::AntennaSetFSR::init","Added to antenna list: "
               +(*antPtrIt)->humanName()+
               " (but one or more consistency checks failed)");
    }
  }
}

//==========================================================================

// AntennaSetISR class.
// This class is a simple container for a set of II & IF antenna functions

void AntennaSetISR::init() {
  
  if (! isInitPtr) {
    printOut("AntennaSetISR::init","Cannot initialize, pointers not set.");
    return;
  }
  
  // Set verbose level
  verbose = settingsPtr->mode("Vincia:verbose");
  
  // Create antenna objects
  antPtrs.resize(12);
  antPtrs[iQQemitII]  = new QQEmitII();
  antPtrs[iGQemitII]  = new GQEmitII();
  antPtrs[iGGemitII]  = new GGEmitII();
  antPtrs[iQXsplitII] = new QXSplitII();
  antPtrs[iGXconvII]  = new GXConvII();
  antPtrs[iQQemitIF]  = new QQEmitIF();
  antPtrs[iQGemitIF]  = new QGEmitIF();
  antPtrs[iGQemitIF]  = new GQEmitIF();
  antPtrs[iGGemitIF]  = new GGEmitIF();
  antPtrs[iQXsplitIF] = new QXSplitIF();
  antPtrs[iGXconvIF]  = new GXConvIF();
  antPtrs[iXGsplitIF] = new XGSplitIF();
  
  // Loop through antPtrs and initialise
  for (vector<AntennaISR*>::iterator antPtrIt = antPtrs.begin(); 
       antPtrIt != antPtrs.end(); ++antPtrIt) {
    
    // Initialize antenna pointers to PYTHIA 8 objects
    (*antPtrIt)->initPtr(functionsPtr, particleDataPtr, rndmPtr, settingsPtr);
    
    // Initialize antenna itself
    bool pass = (*antPtrIt)->init();
    
    // Optionally check antenna (singularities, positivity, etc)
    if (settingsPtr->flag("Vincia:checkAntennae")) 
      pass = pass && (*antPtrIt)->check();
    
    // Everything OK?
    if (pass) {
      if (verbose >= 2)
        printOut("Vincia::AntennaSetISR::init","Added to antenna list: "
                 +(*antPtrIt)->vinciaName());
    } 
    else if (verbose >= 1) {
      printErr("Vincia::AntennaSetISR::init","Added to antenna list: "
               +(*antPtrIt)->vinciaName()+
               " (but one or more consistency checks failed");
    }
  }
}

//==========================================================================

// AntennaFSR class.
// Base class for FF antenna function implementations
  
bool AntennaFSR::init() {
      
  // Check whether pointers are initialized
  if (!isInitPtr) return false;

  // Verbose level
  verbose              = settingsPtr->mode("Vincia:verbose");    
  // Charge factor
  chargeFacSav      = settingsPtr->parm(vinciaName()+":chargeFactor");
  if (chargeFacSav < 0.) chargeFacSav = 0.0;
  // Kinematics map type (first look for specific, else default to global)
  if (settingsPtr->isMode(vinciaName()+":kineMapType"))
    kineMapTypeSav     = settingsPtr->mode(vinciaName()+":kineMapType");
  else
    kineMapTypeSav     = settingsPtr->mode("Vincia:kineMapType");

  // Collinear partitioning (for global antennae)
  alphaSav = settingsPtr->parm("Vincia:octetPartitioning");    

  // Subleading-colour treatment
  modeSLC = settingsPtr->mode("Vincia:modeSLC");
  // modeSLC = 0: all gluon-emission antennae normalised to CA
  if (modeSLC == 0 && idj() == 21) chargeFacSav = CA;
  // modeSLC = 2: 
  if (modeSLC == 2 && idj() == 21) {
    // QG and GG antennae start out normalised to CA.
    // (QG antennae get a phase-space dependent 1/NC^2 correction)
    if ( idA() == 21 || idB() == 21 ) chargeFacSav = CA;
    // QQ antennae are normalised to 2CF (called CF in Vincia's normalisation)
    else chargeFacSav = CF;
  }
    
  // Return OK
  isInit = true;
  return isInit;
    
}
  
//--------------------------------------------------------------------------
  
// The (dimensionless) antenna function. Base class returns zero.

double AntennaFSR::antennaFunction(double, double, 
                                   double, double, double,
                                   int, int, int, int, int) {
  printOut("AntennaFSR::antennaFunction","Error. Dummy function in base "
           "class called.");
  return 0.0;
}

//--------------------------------------------------------------------------

// The AP kernel(s). Base class returns -1.

double AntennaFSR::AltarelliParisi(double, double, double, double, double, 
                                   int, int, int, int, int) {
  return -1.0;
}
  
//--------------------------------------------------------------------------

// Generic check of antenna function for use during initialisation

bool AntennaFSR::check() {

  bool isOK = true;
    
  // Check soft singularity against massless dipole factor
  if (idj() == 21 || idj() == 22) {
    // Test one symmetric and one asymmetric soft phase-space point
    for (int iTest = 0; iTest <= 1; ++iTest) {
      // Test invariants
      double yij = 1.0e-3 * pow(10.0,iTest);
      double yjk = 1.0e-3 / pow(10.0,iTest);
      // Compute eikonal
      double eik = 2.*(1.-yij-yjk)/yij/yjk;
      // Compute antenna (without subleading-colour corrections)
      int modeSLCsave = modeSLC;
      modeSLC = 1;
      double ant = antennaFunction(yij,yjk);
      modeSLC = modeSLCsave;
      // Compare
      double ratio = ant/eik;
      if (abs(ratio-1) >= 0.001) {
        isOK = false;
        if (verbose >= 1) printOut(vinciaName()+":check","WARNING: FAILED "
                                   "(eikonal "+num2str(iTest,1)+")");
      } else if (verbose >= 6) {
        printOut(vinciaName()+":check","OK (eikonal "+
                 num2str(iTest,1)+")");
      }
    }
  }
      
  // Test all helicity configurations for collinearity and posivity
  int hA(9), hB(9), hI(9), hJ(9), hK(9);
  string helString = " 9 9 -> 9 9 9";
  for (int iHel = -1; iHel < 16; ++iHel) {
    if (iHel >= 0) {
      hJ = 2*(iHel%2) - 1;
      hI = 2*((iHel/2)%2) - 1;
      hK = 2*((iHel/4)%2) - 1;
      hA = 2*((iHel/8)%2) - 1;
      hB = 2*((iHel/16)%2) - 1;
      helString = num2str(hA,2) + num2str(hB,2) + " ->"
        + num2str(hI,2) + num2str(hJ,2) + num2str(hK,2);
    }

    // Check collinear singularities against massless AP kernels, if 
    // implemented (base class dummy AP function returns -1)
    for (int iTest=0; iTest<=2; ++iTest) {
      // Test invariants, for a few points along collinear axis.
      double y1   = 1.0e-5;
      double y2   = 0.2 + iTest*0.3;
      // Compute AP kernels
      // V->FF splitting antennae only have JK collinear singularity
      double AP1  = AltarelliParisi(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);
      double AP2  = AltarelliParisi(y2,y1,0.,0.,0.,hA,hB,hI,hJ,hK);
      if ( idj() != 21 && idj() != 22 ) AP1 = -1.;
      // There is only an IJ singularity if hB is conserved 
      if ( hB != hK ) AP1 = -1.;
      if ( hA != hI ) AP2 = -1.;
      // Gluon emission: do not allow external helicity flips (global antennae)
      if ( idj() == 21 || idj() == 22 ) {
	if ( hA != hI || hB != hK ) {
	  AP1 = -1.;
	  AP2 = -1.;
	}
      }
      // Are there any limits to check?
      if (AP1 < 0. && AP2 < 0.) continue;
      
      // Compute antennae (without subleading colour corrections)
      int modeSLCsave = modeSLC;
      modeSLC = 1;
      double ant1 = antennaFunction(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);
      double ant2 = antennaFunction(y2,y1,0.,0.,0.,hA,hB,hI,hJ,hK);      
      // For global antennae, add "swapped" terms to represent neighbor
      if (idA() == 21) {
	ant1 += antennaFunction(y1,1.-y1-y2,0.,0.,0.,hA,hB,hJ,hI,hK);
	ant2 += antennaFunction(y2,1.-y1-y2,0.,0.,0.,hA,hB,hJ,hI,hK);
      }
      if (idB() == 21) {
	ant1 += antennaFunction(1.-y1-y2,y2,0.,0.,0.,hA,hB,hI,hK,hJ);
	ant2 += antennaFunction(1.-y1-y2,y1,0.,0.,0.,hA,hB,hI,hK,hJ);
      }
      // Restore subleading-colour level
      modeSLC = modeSLCsave;
      
      // Compare
      if (AP1 > 0.) {
        double ratio = ant1/AP1;
        if (abs(ratio-1) >= 0.01) {
          isOK = false;
          if (verbose >= 1) 
            printOut(vinciaName()+":check","WARNING (collinear ij "
		     +num2str(iTest,1)+" "+helString+" )");
          if (verbose >= 3) {
            setprecision(6);
            cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                <<" y2 = "<<num2str(y2,9)<<" "<<endl;
            cout<<"    P(z) = "<<num2str(AP1,9)<<"  z = "
                <<num2str(zi(y1,y2),9)<<endl;
          }
        } else if (verbose >= 6) {
          printOut(vinciaName()+":check","OK (collinear ij "
            +num2str(iTest,1)+" "+helString+" )");
        }
      }
      if (AP2 > 0.) {
        double ratio = ant2/AP2;
        if (abs(ratio-1) >= 0.01) {
          isOK = false;
          if (verbose >= 1) 
            printOut(vinciaName()+":check","WARNING: FAILED (collinear jk "
              +num2str(iTest,1)+" "+helString+" )");
          if (verbose >= 3) {
            setprecision(6);
            cout<<"    ant  = "<<num2str(ant2,9)<<" y1 = "<<num2str(y2,9)
                <<" y2 = "<<num2str(y1,9)<<" "<<endl;
            cout<<"    P(z) = "<<num2str(AP2,9)<<"  z = "
                <<num2str(zk(y1,y2),9)<<endl;
          }
        } else if (verbose >= 6) {
          printOut(vinciaName()+":check","OK (collinear jk "
            +num2str(iTest,1)+" "+helString+" )");
        }
      }
    }
    
    // Check positivity
    int nPointsCheck = settingsPtr->mode("Vincia:nPointsCheck");
    bool isPositive  = true;
    bool isZero      = true;
    bool hasDeadZone = false;
    double deadZoneAvoidance = settingsPtr->parm("Vincia:deadZoneAvoidance");
    for (int iTest = 0; iTest < nPointsCheck; ++iTest) {

      // Generate a random point in phase-space triangle
      double y1 = rndmPtr->flat();
      double y2 = rndmPtr->flat();
      if (y1 + y2 > 1.0) {
        y1 = 1 - y1;
        y2 = 1 - y2;
      }
      double ant = antennaFunction(y1,y2,0.,0.,0.,hA,hB,hI,hJ,hK);

      // Check positivity (strict)
      if (ant < 0.) {
        isPositive = false;
        if (verbose >= 3) 
          printOut(vinciaName()+":check","ERROR ant("
                   +num2str(y1,9)+","+num2str(y2,9)+" ; "+helString+") = "
            +num2str(ant)+" < 0");
      } else if (ant > 0.) {
        isZero = false;
      }
      // Check for dead zones away from phase-space boundary
      if (1-y1-y2 > 0.05 && y1 > 0.05 && y2 > 0.05
          && ant < deadZoneAvoidance) hasDeadZone = true;
        
    } // End nPointsCheck sum for positivity checks

      // OK?
    isOK = isOK && isPositive;

    // Verbose output
    if (!isPositive && verbose >= 1) {
      printOut(vinciaName()+":check","ERROR (ant < 0 encountered "
               +helString+" )");
    } else if (isPositive && !isZero && verbose >= 6) {
      printOut(vinciaName()+":check","OK (is positive "
               +helString+" )");
    }
    if (!isZero) {
      if (hasDeadZone && verbose >= 1) {
        printOut(vinciaName()+":check","WARNING (dead zone encountered "
                 +helString+" )");
      } else if (!hasDeadZone && verbose >= 6) {
        printOut(vinciaName()+":check","OK (no dead zones "
                 +helString+" )"); 
      }
    }
      
  } // End helicity sum


  return isOK;
    
}
  
//==========================================================================

// AntennaISR class.
// Base class for II and IF antenna function implementations
  
bool AntennaISR::init() {
      
  // Check whether pointers are initialized
  if (!isInitPtr) return false;
  // Verbose level
  verbose             = settingsPtr->mode("Vincia:verbose");    
  // Charge factor
  chargeFacSav        = settingsPtr->parm(vinciaName()+":chargeFactor");
  if (chargeFacSav < 0.) chargeFacSav = 0.0;

  // Collinear partitioning (for global antennae)
  alphaSav = settingsPtr->parm("Vincia:octetPartitioning");    
    
  // Return OK
  isInit = true;
  return isInit;
    
}
  
//--------------------------------------------------------------------------
  
// The (dimensionless) antenna function. Base class returns zero.

double AntennaISR::antennaFunction(double, double, double, 
                                   int, int, int, int, int) {
  printOut("Antenna::antennaFunction","Error. Dummy function in base "
           "class called.");
  return 0.0;
}

//--------------------------------------------------------------------------

// The AP kernel(s). Base class returns -1.

double AntennaISR::AltarelliParisi(double, double, double, 
                                   int, int, int, int, int) {
  return -1.0;
}

//--------------------------------------------------------------------------

// Generic check of antenna function for use during initialisation

bool AntennaISR::check() {

  bool isOK = true;
      
  // For check for positivity
  int nPointsCheck = settingsPtr->mode("Vincia:nPointsCheck");
  bool isPositive  = true;
  bool isZero      = true;
  bool hasDeadZone = false;
  double deadZoneAvoidance = settingsPtr->parm("Vincia:deadZoneAvoidance");
  // Use 14 TeV as phase space max
  double shhMax    = pow2(14000.0);
    
  // Initial-Initial: sab = sAB + saj + sjb
  if (isIIant()) {
      
    // Z and Higgs as testing case
    double sAB[2] = { pow2(91.0), pow2(125.0) };
    
    // Check soft singularity against dipole factor
    if (idj() == 21 || idj() == 22) {
      // Test one symmetric and one asymmetric soft phase-space point
      for (int iTest = 0; iTest <= 1; ++iTest) {
        // Test two sAB values
        for (int isAB = 0; isAB <= 1; ++isAB) {
          // Test invariants
          double yaj = 1.0e-3 * pow(10.0,iTest); // saj/sab
          double yjb = 1.0e-3 / pow(10.0,iTest); // sjb/sab
          double sab = sAB[isAB]/(1.0-yaj-yjb);  // 1 = sAB/sab + yaj + yjb
          double saj = yaj*sab;
          double sjb = yjb*sab;
          // Compute eikonal
          double eik = 2.0*sab/saj/sjb;
          // Compute antenna
          double ant = antennaFunction(saj,sjb,sAB[isAB]);
          // Compare
          double ratio = ant/eik;
          if (abs(ratio-1.0) >= 0.001) {
            isOK = false;
            if (verbose >= 1)
              printOut(vinciaName()+":check","WARNING: FAILED "
                       "(eikonal "+num2str(iTest,1)+" and sAB = "+
                       num2str((int)sqrt(sAB[isAB]))+"^2)");
          } else if (verbose >= 6) {
            printOut(vinciaName()+":check","OK (eikonal "+num2str(iTest,1)+
                     " and sAB = "+num2str((int)sqrt(sAB[isAB]))+"^2)");
          }
        }
      }
    }
    
    // Check for collinearity (no helicity APs so far)
    // Test invariants, for a few points along collinear axis
    for (int iTest = 0; iTest <= 2; ++iTest) {
      // Test two sAB values
      for (int isAB = 0; isAB <= 1; ++isAB) {
        // Test invariants
        double y1  = 1.0e-5; // saj/sab
        double y2  = 0.2 + iTest*0.3; // sjb/sab
        double sab = sAB[isAB]/(1.0-y1-y2);  // 1 = sAB/sab + yaj + yjb
        double s1  = y1*sab;
        double s2  = y2*sab;
        // Compute AP kernels
        double AP1 = AltarelliParisi(s1,s2,sAB[isAB]);
        double AP2 = AltarelliParisi(s2,s1,sAB[isAB]);
        // Compute antennae
        double ant1 = antennaFunction(s1,s2,sAB[isAB]);
        double ant2 = antennaFunction(s2,s1,sAB[isAB]);
        // Check if only a singularity on one side exists (-> z = -1)
        double zs1 = z1(s1,s2,sAB[isAB]);
        double zs2 = z2(s1,s2,sAB[isAB]);
        if (zs1 != -1.0) {
          double ratio = ant1/AP1;
          if (abs(ratio-1.0) >= 0.01) {
            isOK = false;
            if (verbose >= 1)
              printOut(vinciaName()+":check","WARNING: FAILED "
                       "(collinear aj "+num2str(iTest,1)+" and sAB = "+
                       num2str((int)sqrt(sAB[isAB]))+"^2)");
            if (verbose >= 3) {
              setprecision(6);
              cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                  <<" y2 = "<<num2str(y2,9)<<" "<<endl;
              cout<<"    P(z) = "<<num2str(AP1,9)<<"  z = "
                  <<num2str(zs1,9)<<endl;
            }
          } else if (verbose >= 6) {
            printOut(vinciaName()+":check","OK (collinear aj "+num2str(iTest,1)+
                     " and sAB = "+num2str((int)sqrt(sAB[isAB]))+"^2)");
          }
        }
        if (zs2 != -1.0) {
          double ratio = ant2/AP2;
          if (abs(ratio-1.0) >= 0.01) {
            isOK = false;
            if (verbose >= 1)
              printOut(vinciaName()+":check","WARNING: FAILED "
                       "(collinear jb "+num2str(iTest,1)+" and sAB = "+
                       num2str((int)sqrt(sAB[isAB]))+"^2)");
            if (verbose >= 3) {
              setprecision(6);
              cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                  <<" y2 = "<<num2str(y2,9)<<" "<<endl;
              cout<<"    P(z) = "<<num2str(AP2,9)<<"  z = "
                  <<num2str(zs2,9)<<endl;
            }
          } else if (verbose >= 6) {
            printOut(vinciaName()+":check","OK (collinear jb"+num2str(iTest,1)+
                     " and sAB = "+num2str((int)sqrt(sAB[isAB]))+"^2)");
          }
        }
      }
    }
      
    // Test all helicity configurations for posivity and dead zones
    int hA(9), hB(9), ha(9), hj(9), hb(9);
    string helString = " 9 9 -> 9 9 9";
    for (int iHel = -1; iHel < 0/*16*/; ++iHel) {
      // Begin - loop over hels
      if (iHel >= 0) {
        hj = 2*(iHel%2) - 1;
        ha = 2*((iHel/2)%2) - 1;
        hb = 2*((iHel/4)%2) - 1;
        hA = 2*((iHel/8)%2) - 1;
        hB = 2*((iHel/16)%2) - 1;
        helString = num2str(hA,2) + num2str(hB,2) + " ->" +
          num2str(ha,2) + num2str(hj,2) + num2str(hb,2);
      }
        
      // Gluon emission in global ants: 
      // do not allow external helicity flips
      if ( idj() == 21 || idj() == 22 )
        if ( hA != ha || hB != hb ) continue;
        
      // Check for positivity and dead zones
      for (int iTest = 0; iTest <= nPointsCheck; ++iTest) {
        // sAB^0.5 between 1 and 1000 GeV
        double sABnow = pow2(1.0+rndmPtr->flat()*999.0);
        // Generate a random point in phase-space triangle
        double yaj = rndmPtr->flat(); // saj/sab
        double yjb = rndmPtr->flat(); // sjb/sab
        if (yaj + yjb > 1.0) {
          yaj = 1.0-yaj;
          yjb = 1.0-yjb;
        }
        // Compute sab from 1 = sAB/sab + yaj + yjb
        double sab = sABnow/(1.0-yaj-yjb);  
        // Check that sab < shhMax
        if (sab > shhMax) continue;
        double saj = yaj*sab;
        double sjb = yjb*sab;
        // Compute antenna
        double ant = antennaFunction(saj,sjb,sABnow,hA,hB,ha,hj,hb);
        // Check positivity (strict)
        if (ant < 0.0) {
          isPositive = false;
          if (verbose >= 3) 
            printOut(vinciaName()+":check","ERROR ant("+num2str((int)sqrt(saj))
                     +"^2,"+num2str((int)sqrt(sjb))+"^2,"
                     +num2str((int)sqrt(sABnow))+"^2 ; "+helString+") < 0");
        } else if (ant > 0.0) {
          isZero = false;
        }
        // Check for dead zones away from phase-space boundary
        if ( (1.0-yaj-yjb > 0.05) && (yaj > 0.05) && (yjb > 0.05)
             && (ant*sABnow < deadZoneAvoidance) ) hasDeadZone = true;
      }
      
      // OK?
      isOK = isOK && isPositive;
        
      // Verbose output
      if (!isPositive && verbose >= 1) {
        printOut(vinciaName()+":check","ERROR (ant < 0 encountered "
                 +helString+" )");
      } else if (isPositive && !isZero && verbose >= 6) {
        printOut(vinciaName()+":check","OK (is positive "
                 +helString+" )");
      }
      if (!isZero) {
        if (hasDeadZone && verbose >= 1) {
          printOut(vinciaName()+":check","WARNING (dead zone encountered "
                   +helString+" )");
        } else if (!hasDeadZone && verbose >= 6) {
          printOut(vinciaName()+":check","OK (no dead zones "
                   +helString+" )");
        }
      }
    } // End - loop over hels
      
  }
  // Initial-Initial
      
  // Initial-Final: sAK + sjk = sak + saj
  else {
      
    // Two values for sAK
    double sAK[2] = { pow2(50.0), pow2(150.0) };
    
    // Check soft singularity against dipole factor
    if (idj() == 21 || idj() == 22) {
      // Test one symmetric and one asymmetric soft phase-space point
      for (int iTest = 0; iTest <= 1; ++iTest) {
        // Test two sAK values
        for (int isAK = 0; isAK <= 1; ++isAK) {
          // Test invariants
          double yaj = 1.0e-3 * pow(10.0,iTest); // saj/sAK
          double yjk = 1.0e-3 / pow(10.0,iTest); // sjk/sAK
          double sak = sAK[isAK]*(1.0+yjk-yaj); // 1 + yjk = sak/sAK + yaj
          double saj = yaj*sAK[isAK];
          double sjk = yjk*sAK[isAK];
          // Compute eikonal
          double eik = 2.0*sak/saj/sjk;
          // Compute antenna
          double ant = antennaFunction(saj,sjk,sAK[isAK]);
          // Compare
          double ratio = ant/eik;
          if (abs(ratio-1.0) >= 0.001) {
            isOK = false;
            if (verbose >= 1)
              printOut(vinciaName()+":check","WARNING: FAILED "
                       "(eikonal "+num2str(iTest,1)+" and sAK = "+
                       num2str((int)sqrt(sAK[isAK]))+"^2)");
          } else if (verbose >= 6) {
            printOut(vinciaName()+":check","OK (eikonal "+num2str(iTest,1)+
                     " and sAK = "+num2str((int)sqrt(sAK[isAK]))+"^2)");
          }
        }
      }
    }
    
    // Check for collinearity (no helicity APs so far)
    // Test invariants, for a few points along collinear axis
    for (int iTest = 0; iTest <= 2; ++iTest) {
      // Test two sAK values
      for (int isAK = 0; isAK <= 1; ++isAK) {
        // Test invariants
        double y1  = 1.0e-5; // saj/sAK
        double y2  = 0.2 + iTest*0.3; // sjk/sAK
        double s1  = y1*sAK[isAK];
        double s2  = y2*sAK[isAK];
        // Compute AP kernels
        double AP1 = AltarelliParisi(s1,s2,sAK[isAK]);
        double AP2 = AltarelliParisi(s2,s1,sAK[isAK]);
        // Compute antennae
        double ant1 = antennaFunction(s1,s2,sAK[isAK]);
        double ant2 = antennaFunction(s2,s1,sAK[isAK]);
        // Check if only a singularity on one side exists (-> z = -1)
        double zs1 = z1(s1,s2,sAK[isAK]);
        double zs2 = z2(s1,s2,sAK[isAK]);
        if (zs1 != -1.0) {
          double ratio = ant1/AP1;
          if (abs(ratio-1.0) >= 0.01) {
            isOK = false;
            if (verbose >= 1)
              printOut(vinciaName()+":check","WARNING: FAILED "
                       "(collinear aj "+num2str(iTest,1)+" and sAK = "+
                       num2str((int)sqrt(sAK[isAK]))+"^2)");
            if (verbose >= 3) {
              setprecision(6);
              cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                  <<" y2 = "<<num2str(y2,9)<<" "<<endl;
              cout<<"    P(z) = "<<num2str(AP1,9)<<"  z = "
                  <<num2str(zs1,9)<<endl;
            }
          } else if (verbose >= 6) {
            printOut(vinciaName()+":check","OK (collinear aj "+num2str(iTest,1)
                     +" and sAK = "+num2str((int)sqrt(sAK[isAK]))+"^2)");
          }
        }
        if (zs2 != -1.0) {
          double ratio = ant2/AP2;
          if (abs(ratio-1.0) >= 0.01) {
            isOK = false;
            if (verbose >= 1)
              printOut(vinciaName()+":check","WARNING: FAILED "
                       "(collinear jk "+num2str(iTest,1)+" and sAK = "+
                       num2str((int)sqrt(sAK[isAK]))+"^2)");
            if (verbose >= 3) {
              setprecision(6);
              cout<<"    ant  = "<<num2str(ant1,9)<<" y1 = "<<num2str(y1,9)
                  <<" y2 = "<<num2str(y2,9)<<" "<<endl;
              cout<<"    P(z) = "<<num2str(AP2,9)<<"  z = "
                  <<num2str(zs2,9)<<endl;
            }
          } else if (verbose >= 6) {
            printOut(vinciaName()+":check",
                     "OK (collinear jk "+num2str(iTest,1)+
                     " and sAK = "+num2str((int)sqrt(sAK[isAK]))+"^2)");
          }
        }
      }
    }
      
    // Test all helicity configurations for posivity and dead zones
    int hA(9), hK(9), ha(9), hj(9), hk(9);
    string helString = " 9 9 -> 9 9 9";
    for (int iHel = -1; iHel < 0/*16*/; ++iHel) { // Begin - loop over hels
      if (iHel >= 0) {
        hj = 2*(iHel%2) - 1;
        ha = 2*((iHel/2)%2) - 1;
        hk = 2*((iHel/4)%2) - 1;
        hA = 2*((iHel/8)%2) - 1;
        hK = 2*((iHel/16)%2) - 1;
        helString = num2str(hA,2) + num2str(hK,2) + " ->" +
          num2str(ha,2) + num2str(hj,2) + num2str(hk,2);
      }
        
      // Gluon emission: do not allow external helicity flips
      if ( idj() == 21 || idj() == 22 )
        if ( hA != ha || hK != hk ) continue;
        
      // Check for positivity and dead zones
      for (int iTest = 0; iTest <= nPointsCheck; ++iTest) {
        // sAK^0.5 between 1 and 1000 GeV
        double sAKnow = pow2(1.0+rndmPtr->flat()*999.0);
        // pick random xA
        double xA  = rndmPtr->flat();
        // sjk between 0 and sAK(1-xA)/xA, saj between 0 and sAK+sjk
        double sjk = rndmPtr->flat()*sAKnow*(1.0-xA)/xA;
        double saj = rndmPtr->flat()*(sAKnow+sjk);
        double sak = sAKnow+sjk-saj;
        // Check that sxy < shhMax
        if (saj > shhMax || sjk > shhMax || sak > shhMax || sAKnow > shhMax)
          continue;
        // Compute antenna
        double ant = antennaFunction(saj,sjk,sAKnow,hA,hK,ha,hj,hk);
        // Check positivity (strict)
        if (ant < 0.0) {
	  cout << "ant = "<<ant * sAKnow <<endl;
          isPositive = false;
          if (verbose >= 3) 
            printOut(vinciaName()+":check","ERROR ant("+num2str((int)sqrt(saj))
                     +"^2,"+num2str((int)sqrt(sjk))+"^2,"
                     +num2str((int)sqrt(sAKnow))+"^2 ; "+helString+") < 0");
        } else if (ant > 0.0) {
          isZero = false;
        }
        // Check for dead zones away from phase-space boundary
        if ( (sjk < 0.95*sAKnow*(1.0-xA)/xA) && (sjk > 1.0) && (saj > 1.0)
             && (sak > 1.0) && (ant*sAKnow < deadZoneAvoidance) )
          hasDeadZone = true;
      }
      
      // OK?
      isOK = isOK && isPositive;
        
      // Verbose output
      if (!isPositive && verbose >= 1) {
        printOut(vinciaName()+":check","ERROR (ant < 0 encountered "
                 +helString+" )");
      } else if (isPositive && !isZero && verbose >= 6) {
        printOut(vinciaName()+":check","OK (is positive "
                 +helString+" )");
      }
      if (!isZero) {
        if (hasDeadZone && verbose >= 1) {
          printOut(vinciaName()+":check","WARNING (dead zone encountered "
                   +helString+" )");
        } else if (!hasDeadZone && verbose >= 6) {
          printOut(vinciaName()+":check","OK (no dead zones "
                   +helString+" )");
        }
      }
    } // End - loop over hels
      
  }
  // Initial-Final

  return isOK;
    
}

//==========================================================================
//
// Final-Final Antenna Functions A B -> i j k
// Helicity-dependent only if all partons massless
// 
//==========================================================================

// QQEmitFF
// Dimensionless parameters: yij = 2pi.pj / m2ijk; mui = mi / mijk
  
double QQEmitFF::antennaFunction(double yij, double yjk, 
                               double mui, double muj, double muk,
                               int hA, int hB, int hi, int hj, int hk) {
    
  // Sanity check. Require positive invariants.
  if (yij <= 0.0 || yjk <= 0.0) return 0.0;

  // No clue about massive gluons
  if (muj != 0.) {
    if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                               "Ignoring gluon virtuality");
    muj = 0.0;
  }

  // Massive partons are always treated as unpolarised: 
  if (mui > 0.) {
    hi = 9;
    hA = 9;
  } else {
    mui = 0.;
  }
  if (muk > 0.) {
    hk = 9;
    hB = 9;
  } else {
    muk = 0.;
  }
    
  // Treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( hi != 1 && hi != -1 ) hi = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;

  // Check physical helicity combination
  if ( hi == -hA ) return 0.0;
  else if ( hk == -hB ) return 0.0;
  
  // How many helicity configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthand
  double eik      = 1./yij/yjk;
  
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
    // ++ > +++ && -- > --- antennae (MHV)
    term = eik;
    if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV)
    term = eik * pow2(1.-yij-yjk);
    if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
    // +- > ++- && -+ > --+ antennae
    term = eik * pow2(1.-yij);
    if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = eik * pow2(1.-yjk);
    if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;
  }

  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
    
}

//==========================================================================

// QGEmitFF
// Dimensionless parameters: yij = 2pi.pj / m2ijk; mui = mi / mijk

double QGEmitFF::antennaFunction(double yij, double yjk, 
                               double mui, double muj, double muk,
                               int hA, int hB, int hi, int hj, int hk) {
    
  // Sanity check. Require positive invariants.
  if (yij <= 0.0 || yjk <= 0.0) return 0.0;

  // No clue about massive gluons
  if (muj != 0. || muk != 0.) {
    if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                               "Ignoring gluon virtuality");
    muj = 0.0;
    muk = 0.0;
  }

  // Massive partons are always treated as unpolarised: 
  if (mui > 0.) {
    hi = 9;
    hA = 9;
  } else {
    mui = 0.;
  }
    
  // Treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( hi != 1 && hi != -1 ) hi = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // Check physical helicity combination
  if ( hi == -hA ) return 0.0;
  else if ( hk == -hB ) return 0.0;
  
  // Hhow many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double yik      = max(0.,1. - yij - yjk);
  double eik      = 1./yij/yjk;
  double a        = 1.-alphaSav;
  if (alphaSav == 0.0) a = 1.;
  else if (alphaSav == 1.0) a = 0.;
  
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
    // ++ > +++ && -- > --- antennae (MHV)
    term = eik;
    if (a != 0.) term += a * (1.-yjk) * ( 1. - 2*yij - yjk )/yjk;
    if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV) 
    term = eik * pow2(yik) * (1.-yij);
    if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
    // +- > ++- && -+ > --+ antennae
    term = eik * pow3(1.-yij);
    if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = eik * pow2(1.-yjk);
    if (a != 0.) term += a * (1.-yjk) * ( 1. - 2*yij - yjk )/yjk;
    if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;

  }

  // Subleading colour corrections
  if (modeSLC >= 2) {
    double colFac = 1./pow2(NC);
    // + > ++ and - > --   (with finite terms to help make hSum > 0)
    term =  - colFac * ( ( 1./yij/yjk ) * yjk / ( yij + yjk ) - ( 1. + yjk ) );
    if (hi == hj || hj == 9) hSum += term;
    // + > +- and - > -+   (with finite terms to add up to QQ antenna)
    if (hj != hi || hj == 9) {
      term += -colFac * ( ( -2. + yjk ) / yij + ( 1. + yjk ) + 2. );
      hSum += term;
    }
    // For very small total antenna values, sum can be < 0; force nonnegative
    // (E.g., in our standard parametrisation, the helicity-flip gluon side
    // goes to zero for hard-collinear.) 
    hSum = max(0.,hSum);
  }

  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
    
}

//==========================================================================

// GGEmitFF
// Dimensionless parameters: yij = 2pi.pj / m2ijk; mui = mi / mijk
  
double GGEmitFF::antennaFunction(double yij, double yjk, 
                               double mui, double muj, double muk,
                               int hA, int hB, int hi, int hj, int hk) {
    
  // Sanity check. Require positive invariants.
  if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
  // No clue about massive gluons
  if (mui != 0. || muj != 0. || muk != 0.) {
    if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                               "Ignoring gluon virtuality.");
    mui = 0.0;
    muj = 0.0;
    muk = 0.0;
  }
    
  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( hi != 1 && hi != -1 ) hi = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // Check physical helicity combination
  if ( hi == -hA ) return 0.0;
  else if ( hk == -hB ) return 0.0;
  
  // Hhow many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double eik      = 1./yij/yjk;
  double yik      = max(0.,1. - yij - yjk);
  double a;
  if (alphaSav == 0.0) a = 1.;
  else if (alphaSav == 1.0) a = 0.;
  else a = 1.-alphaSav;

  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
    // ++ > +++ && -- > --- antennae (MHV)
    term = eik;
    if (a != 0.0) term += a * (   (1.-yjk)*(1.-2*yij-yjk)/yjk
				+ (1.-yij)*(1.-2*yjk-yij)/yij );
    if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && LH[hk] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV)
    term = eik * pow3(yik);
    if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += term;

  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
    // +- > ++- && -+ > --+ antennae
    term = eik * pow3(1-yij);
    if (a != 0.) term += a * (1.-yij)*(1.-2*yjk)/yij;
    if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = eik * pow3(1-yjk);
    if (a != 0.) term += a * (1.-yjk)*(1.-2*yij)/yjk;
    if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && RH[hk] ) hSum += term;

  }

  // Return helicity sum, averaged over initial helicities 
  return hSum/nAvg;
    
}

//==========================================================================

// QGSplitFF
// Dimensionless parameters: yij = 2pi.pj / m2ijk; mui = mi / mijk

double QGSplitFF::antennaFunction(double yij, double yjk, 
                                double mui, double muj, double muk,
                                int hA, int hB, int hi, int hj, int hk) {
    
  // Sanity check. Require positive invariants.
  if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
  // Require equal jk masses
  if ( abs(muj-muk) > TINY ) {
    if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                               "Warning: forcing mq = mqbar.");
    cout<<"QGSPLIT: y(ij,jk) = "<<yij<<" "<<yjk<<" mu(i,j,k) = "
        <<scientific<<mui<<" "<<muj<<" "<<muk<<endl;
    muk = muj;
  }
    
  // Massive partons are always treated as unpolarised: 
  if (mui > 0.) {
    hi = 9;
    hA = 9;
  } else {
    mui = 0.;
  }
  if (muj > 0.) hj = 9;
  else muj = 0.;
  if (muk > 0.) hk = 9;
  else muk = 0.;

  // Treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( hi != 1 && hi != -1 ) hi = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
  
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double mu2i  = pow2(mui);
  double mu2q  = pow2(muj);
  double yik   = 1. - yij - yjk - 2*mu2q - mu2i;
  double m2jk  = yjk + 2 * mu2q;
  double xj    = 1. - yik / (1. - mu2i);
  double xk    = 1. - yij / (1. - mu2i);
  double termQ = 0.5 / m2jk * pow2(1. - xj);
  double termA = 0.5 / m2jk * pow2(1. - xk);
    
  // Do helicity sum
  double hSum = 0.0;
    
  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
    // ++ > +-+ && -- > -+- antennae (quark takes gluon helicity)
    if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += termQ;
    if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += termQ;
      
    // ++ > ++- && -- > --+ antennae (antiquark takes gluon helicity)
    if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += termA;
    if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += termA;
      
  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
    // +- > ++- && -+ > --+ antennae (quark takes gluon helicity)
    if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += termQ;
    if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += termQ;
      
    // +- > +-+ && -+ > -+- antennae (antiquark takes gluon helicity)
    if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += termA;
    if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += termA;
      
  }
        
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}
  
//--------------------------------------------------------------------------
  
//==========================================================================
  
// GGSplitFF : A(g) B(g) -> i(g) j(qbar) k(q)
// Dimensionless parameters: yij = 2pi.pj / m2ijk; mui = mi / mijk
  
double GGSplitFF::antennaFunction(double yij, double yjk, 
                                double mui, double muj, double muk,
                                int hA, int hB, int hi, int hj, int hk) {
    
  // Sanity check. Require positive invariants.
  if (yij <= 0.0 || yjk <= 0.0) return 0.0;
    
  // Ignore gluon mass
  if (mui != 0.0) {
    if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                               "Ignoring gluon virtuality");
    mui = 0.0;
  }
    
  // Require equal jk masses (if no k given, set k = j)
  if ( abs(muj-muk) > TINY ) {
    if (verbose >= 2) printOut(vinciaName()+":antennaFunction",
                               "Warning: forcing mq = mqbar.");
    muj = max(0.,(muj + muk)/2.);
    muk = muj;
  }

  // Massive partons are always treated as unpolarised: 
  if (muj > 0.) hj = 9;
  else muj = 0.;
  if (muk > 0.) hk = 9;
  else muk = 0.;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( hi != 1 && hi != -1 ) hi = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double mu2q  = pow2(muj);
  double yik   = 1. - yij - yjk - 2*mu2q;
  double m2jk  = yjk + 2 * mu2q;
  double xj    = 1. - yik;
  double xk    = 1. - yij;
  double termQ = 0.5 / m2jk * pow2(1. - xj);
  double termA = 0.5 / m2jk * pow2(1. - xk);

  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
      
    // ++ > +-+ && -- > -+- antennae (quark takes gluon helicity)
    if ( RH[hA] && RH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += termQ;
    if ( LH[hA] && LH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += termQ;

    // ++ > ++- && -- > --+ antennae (antiquark takes gluon helicity)
    if ( RH[hA] && RH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += termA;
    if ( LH[hA] && LH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += termA;
  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
      
    // +- > ++- && -+ > --+ antennae (quark takes gluon helicity)
    if ( RH[hA] && LH[hB] && RH[hi] && RH[hj] && LH[hk] ) hSum += termQ;
    if ( LH[hA] && RH[hB] && LH[hi] && LH[hj] && RH[hk] ) hSum += termQ;
      
    // +- > +-+ && -+ > -+- antennae (antiquark takes gluon helicity)
    if ( RH[hA] && LH[hB] && RH[hi] && LH[hj] && RH[hk] ) hSum += termA;
    if ( LH[hA] && RH[hB] && LH[hi] && RH[hj] && LH[hk] ) hSum += termA;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//==========================================================================
//
// Initial-Initial Antenna Functions A B -> a j b
// 
//==========================================================================

// QQEmitII
// Note: dimensionful parameters [s] = GeV^2

double QQEmitII::antennaFunction(double saj, double sjb, double sAB,
                                 int hA, int hB, int ha, int hj, int hb) {

  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hb != 1 && hb != -1 ) hb = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double sab      = saj+sjb+sAB;
  double yaj      = saj/sab;
  double yjb      = sjb/sab;
  double yAB      = sAB/sab;
  double eikJ     = 1./(sAB*yaj*yjb);
    
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
    // ++ > +++ && -- > --- antennae (MHV)
    term = eikJ;
    if ( RH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term; 
    if ( LH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV)
    term = pow2(yAB)*eikJ;
    if ( RH[hA] && RH[hB] && RH[ha] && LH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[ha] && RH[hj] && LH[hb] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
    // +- > ++- && -+ > --+ antennae
    term = pow2(1.-yaj)*eikJ;
    if ( RH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = pow2(1.-yjb)*eikJ;
    if ( RH[hA] && LH[hB] && RH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && RH[hj] && RH[hb] ) hSum += term;
  }

  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg ;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double QQEmitII::AltarelliParisi(double saj, double sjb, double sAB, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
  // Using smaller invariant for collinear limit
  double z  = ( saj < sjb ? z1(saj,sjb,sAB) : z2(saj,sjb,sAB) );
  double Q2 = min(saj,sjb);
  double AP = 1.0/z * (1.0+pow2(z))/(1.0-z);
  // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
  AP *= 1.0;
  // ant with quark reproduces full AP
  AP *= 1.0;
        
  return AP/Q2;
}

//==========================================================================

// GQEmitII
// Note: dimensionful parameters [s] = GeV^2

double GQEmitII::antennaFunction(double saj, double sjb, double sAB, 
                                 int hA, int hB, int ha, int hj, int hb) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hb != 1 && hb != -1 ) hb = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double sab       = saj+sjb+sAB;
  double yaj       = saj/sab;
  double yjb       = sjb/sab;
  double yAB       = sAB/sab;
  double eikJ    = 1./(sAB*yaj*yjb);
  double eikA    = 1./(sAB*yaj*(yAB+yaj));
    
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
    // ++ > +++ && -- > --- antennae (MHV)
    term = eikJ + eikA;
    if ( RH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term; 
    if ( LH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV)
    term = (1.-yjb)*pow2(yAB)*eikJ;
    if ( RH[hA] && RH[hB] && RH[ha] && LH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[ha] && RH[hj] && LH[hb] ) hSum += term;
      
    // ++ > --+ && -- > ++- antennae
    term = pow3(yjb)*eikA;
    if ( RH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
    // +- > ++- && -+ > --+ antennae
    term = pow3(1.-yjb)*eikJ;
    if ( RH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = pow2(1.-yaj)*eikJ + eikA;
    if ( RH[hA] && LH[hB] && RH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && RH[hj] && RH[hb] ) hSum += term;
      
    // +- > --- && -+ > +++ antennae
    term = pow3(yjb)*eikA;
    if ( RH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term;
  }
        
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double GQEmitII::AltarelliParisi(double saj, double sjb, double sAB, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
  // Using smaller invariant for collinear limit
  double z  = ( saj < sjb ? z1(saj,sjb,sAB) : z2(saj,sjb,sAB) );
  double Q2 = min(saj,sjb);
  double AP = 0.0;
    
  if (saj < sjb) { // collinear to initial gluon
    AP = 1.0/z*(pow(z,4.0)+1.0+pow(1-z,4.0))/z/(1.0-z);
    // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
    AP *= 2.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
  }
  else { // collinear to initial quark
    AP = 1.0/z * (1.0+pow2(z))/(1.0-z);
    // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
    AP *= 1.0;
    // ant with quark reproduces full AP
    AP *= 1.0;
  }
        
  return AP/Q2;
}

//==========================================================================

// GGEmitII
// Note: dimensionful parameters [s] = GeV^2

double GGEmitII::antennaFunction(double saj, double sjb, double sAB, 
                                 int hA, int hB, int ha, int hj, int hb) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hb != 1 && hb != -1 ) hb = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double sab     = saj+sjb+sAB;
  double yaj     = saj/sab;
  double yjb     = sjb/sab;
  double yAB     = sAB/sab;
  double eikJ    = 1./(sAB*yaj*yjb);
  double eikA    = 1./(sAB*yaj*(yAB+yaj));
  double eikB    = 1./(sAB*yjb*(yAB+yjb));
    
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
    // ++ > +++ && -- > --- antennae (MHV)
    term = eikJ + eikA + eikB;
    if ( RH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV)
    term = pow3(yAB)*eikJ;
    if ( RH[hA] && RH[hB] && RH[ha] && LH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[ha] && RH[hj] && LH[hb] ) hSum += term;
      
    // ++ > --+ && -- > ++- antennae
    term = pow3(yjb)*eikA;
    if ( RH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term;
      
    // ++ > +-- && -- > -++ antennae
    term = pow3(yaj)*eikB;
    if ( RH[hA] && RH[hB] && RH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[ha] && RH[hj] && RH[hb] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
    // +- > ++- && -+ > --+ antennae
    term = pow3(1.-yaj)*eikJ + eikA;
    if ( RH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = pow3(1-yjb)*eikJ + eikB;
    if ( RH[hA] && LH[hB] && RH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && RH[hj] && RH[hb] ) hSum += term;

    // +- > +++ && -+ > --- antennae
    term = pow3(yaj)*eikB;
    if ( RH[hA] && LH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
      
    // +- > --- && -+ > +++ antennae
    term = pow3(yjb)*eikA;
    if ( RH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double GGEmitII::AltarelliParisi(double saj, double sjb, double sAB, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
  // Using smaller invariant for collinear limit
  double z  = ( saj < sjb ? z1(saj,sjb,sAB) : z2(saj,sjb,sAB) );
  double Q2 = min(saj,sjb);
  double AP = 1.0/z*(pow(z,4.0)+1.0+pow(1-z,4.0))/z/(1.0-z);
  // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
  AP *= 2.0;
  // gluon appears in 2 ants, hence ant reproduces half of AP
  AP *= 0.5;
        
  return AP/Q2;
}

//==========================================================================

// QXSplitII 
// Note: dimensionful parameters [s] = GeV^2
// Note: Splitting in the forwards sense aka quark backwards evolving
//       to a gluon and emitting a antiquark in the final state

double QXSplitII::antennaFunction(double saj, double sjb, double sAB, 
                                  int hA, int hB, int ha, int hj, int hb) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hb != 1 && hb != -1 ) hb = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double sab = sAB + saj + sjb;
  double yaj = saj/sab;
  double yAB = sAB/sab;
  double splitA  = 1.0/sAB/yaj;

  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
    // ++ > +-+ && -- > -+- antennae
    term = pow2(yAB) * splitA;
    if ( RH[hA] && RH[hB] && RH[ha] && LH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[ha] && RH[hj] && LH[hb] ) hSum += term;
      
    // ++ > --+ && -- > ++- antennae
    term = pow2(1.-yAB) * splitA;
    if ( RH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
    // +- > +-- && -+ > -++ antennae
    term = pow2(yAB) * splitA;
    if ( RH[hA] && LH[hB] && RH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && RH[hj] && RH[hb] ) hSum += term;
      
    // +- > --- && -+ > +++ antennae
    term = pow2(1.-yAB) * splitA;
    if ( RH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term;
  }
        
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double QXSplitII::AltarelliParisi(double saj, double sjb, double sAB, 
                                  int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
  // There is only the saj collinear limit
  double z  = z1(saj,sjb,sAB);
  double Q2 = saj;
  double AP = 1.0/z * (pow2(z)+pow2(1-z));
  // AP uses alpha/2pi * TR/2, ant uses alpha/4pi * TR, with TR = 1
  AP *= 1.0;
  // ant with quark reproduces full AP
  AP *= 1.0;
        
  return AP/Q2;
}

//==========================================================================

// GXConvII 
// Note: dimensionful parameters [s] = GeV^2
// Note: Gluon evolves backwards into an IS quark and emits an FS quark

double GXConvII::antennaFunction(double saj, double sjb, double sAB, 
                                 int hA, int hB, int ha, int hj, int hb) {

  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hB != 1 && hB != -1 ) hB = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hb != 1 && hb != -1 ) hb = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hB == 9) nAvg *= 2;
    
  // Shorthands
  double sab     = sAB + saj + sjb;
  double yaj     = saj/sab;
  double yAB     = sAB/sab;
  double convA   = 1.0/(2.*sAB*yaj*(yAB+yaj));

  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hB > 0 || hA == 9 || hB == 9) {
    // ++ > +++ && -- > --- antennae
    term = convA;
    if ( RH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
      
    // ++ > --+ && -- > ++- antennae
    term = pow2(1-yAB)*convA;
    if ( RH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term;
    if ( LH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hB < 0 || hA == 9 || hB == 9) {
    // +- > --- && -+ > +++ antennae
    term = convA;
    if ( RH[hA] && LH[hB] && LH[ha] && LH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && RH[ha] && RH[hj] && RH[hb] ) hSum += term;
      
    // +- > ++- && -+ > --+ antennae
    term = pow2(1-yAB)*convA;
    if ( RH[hA] && LH[hB] && RH[ha] && RH[hj] && LH[hb] ) hSum += term;
    if ( LH[hA] && RH[hB] && LH[ha] && LH[hj] && RH[hb] ) hSum += term;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double GXConvII::AltarelliParisi(double saj, double sjb, double sAB, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjb <= 0.0) || (sAB <= 0.0)) return 0.0;
    
  // There is only the saj collinear limit
  double z  = z1(saj,sjb,sAB);
  double Q2 = saj;
  double AP = 1.0/z * (1.0+pow2(1.0-z))/z;
  // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
  AP *= 1.0;
  // gluon appears in 2 ants, hence ant reproduces half of AP
  AP *= 0.5;
        
  return AP/Q2;
}

//==========================================================================
//
// Initial-Final Antenna Functions A K -> a j k
// Helicity-dependent only if all partons massless
// 
//==========================================================================

// QQEmitIF
// Note: dimensionful parameters [s] = GeV^2

double QQEmitIF::antennaFunction(double saj, double sjk, double sAK, 
                                 int hA, int hK, int ha, int hj, int hk) {

  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hK != 1 && hK != -1 ) hK = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hK == 9) nAvg *= 2;
    
  // Shorthands
  double s         = sAK + sjk;
  double yaj       = saj/s;
  double yjk       = sjk/s;
  double eikJ      = 1./(sAK*yaj*yjk);
  
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hK > 0 || hA == 9 || hK == 9) {
    // ++ > +++ && -- > --- antennae (MHV)
    term = eikJ;
    if ( RH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV)
    term = (pow2(1.-yaj) + pow2(1.-yjk) - 1.) * eikJ
      + ( 3. - pow2(yaj) - pow2(yjk) )/sAK;
    if ( RH[hA] && RH[hK] && RH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && LH[ha] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hK < 0 || hA == 9 || hK == 9) {
    term = pow2(1.-yaj) * eikJ;
    // +- > ++- && -+ > --+ antennae
    if ( RH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = pow2(1.-yjk) * eikJ;
    if ( RH[hA] && LH[hK] && RH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && RH[hj] && RH[hk] ) hSum += term;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double QQEmitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
  // Using smaller invariant for collinear limit
  double z  = ( saj < sjk ? z1(saj,sjk,sAK) : z2(saj,sjk,sAK) );
  double Q2 = min(saj,sjk);
  double AP = 0.0;
    
  if (saj < sjk) { // collinear to initial quark
    AP = 1.0/z * (1.0+pow2(z))/(1.0-z);
    // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
    AP *= 1.0;
    // ant with quark reproduces full AP
    AP *= 1.0;
  }
  else { // collinear to final quark
    AP = (1.0+pow2(z))/(1.0-z);
    // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
    AP *= 1.0;
    // ant with quark reproduces full AP
    AP *= 1.0;
  }
        
  return AP/Q2;
}

//==========================================================================

// QGEmitIF
// Note: dimensionful parameters [s] = GeV^2

double QGEmitIF::antennaFunction(double saj, double sjk, double sAK, 
                                 int hA, int hK, int ha, int hj, int hk) {

  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hK != 1 && hK != -1 ) hK = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hK == 9) nAvg *= 2;
    
  // Shorthands
  double s         = sAK + sjk;
  double yaj       = saj/s;
  double yjk       = sjk/s;
  double eikJ      = 1./(sAK*yaj*yjk);
  double colK      = (1.-alphaSav) * (1.-2.*yaj) / (sAK * yjk);  

  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hK > 0 || hA == 9 || hK == 9) {
    // ++ > +++ && -- > --- antennae (MHV)
    term = eikJ + colK;
    if ( RH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV) 
    term = (pow3(1.-yaj)+pow2(1.-yjk)-1) * eikJ + (3.-pow2(yaj))/sAK;
    if ( RH[hA] && RH[hK] && RH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && LH[ha] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hK < 0 || hA == 9 || hK == 9) {
    // +- > ++- && -+ > --+ antennae
    term = pow3(1.-yaj) * eikJ;
    if ( RH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = pow2(1.-yjk) * eikJ + colK + (2*yaj - yjk)/sAK;
    if ( RH[hA] && LH[hK] && RH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && RH[hj] && RH[hk] ) hSum += term;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double QGEmitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
  // Using smaller invariant for collinear limit
  double z  = ( saj < sjk ? z1(saj,sjk,sAK) : z2(saj,sjk,sAK) );
  double Q2 = min(saj,sjk);
  double AP = 0.0;
    
  if (saj < sjk) { // collinear to initial quark
    AP = 1.0/z * (1.0+pow2(z))/(1.0-z);
    // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
    AP *= 1.0;
    // ant with quark reproduces full AP
    AP *= 1.0;
  }
  else { // collinear to final gluon
    AP = (2.0*z/(1.0-z)+z*(1.0-z));
    // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
    AP *= 2.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
  }
        
  return AP/Q2;
}

//==========================================================================

// GQEmitIF
// Note: dimensionful parameters [s] = GeV^2

double GQEmitIF::antennaFunction(double saj, double sjk, double sAK, 
                                 int hA, int hK, int ha, int hj, int hk) {

  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hK != 1 && hK != -1 ) hK = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hK == 9) nAvg *= 2;
    
  // Shorthands
  double s         = sAK + sjk;
  double yaj       = saj/s;
  double yjk       = sjk/s;
  double yAK       = sAK/s;
  double eikJ      = 1./(sAK*yaj*yjk);
  double eikA      = 1./(sAK*yaj*(yAK+yaj));
    
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hK > 0 || hA == 9 || hK == 9) {

    // ++ > +++ && -- > --- antennae (MHV)
    term = eikJ + eikA;
    if ( RH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;

    // ++ > +-+ && -- > -+- antennae (NMHV)
    term = (pow2(1.-yaj)+pow3(1.-yjk)-1)* eikJ + (3.-pow2(yjk))/sAK;
    if ( RH[hA] && RH[hK] && RH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && LH[ha] && RH[hj] && LH[hk] ) hSum += term;
      
    // ++ > --+ && -- > ++- antennae
    term = pow3(yjk) * eikA;
    if ( RH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hK < 0 || hA == 9 || hK == 9) {
    // +- > ++- && -+ > --+ antennae
    term = pow2(1-yaj) * eikJ + eikA;
    if ( RH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = pow3(1.-yjk) * eikJ;
    if ( RH[hA] && LH[hK] && RH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && RH[hj] && RH[hk] ) hSum += term;
      
    // +- > --- && -+ > +++ antennae
    term = pow3(yjk) * eikA;
    if ( RH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term;
  }    
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double GQEmitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
  // Using smaller invariant for collinear limit
  double z  = ( saj < sjk ? z1(saj,sjk,sAK) : z2(saj,sjk,sAK) );
  double Q2 = min(saj,sjk);
  double AP = 0.0;
    
  if (saj < sjk) { // collinear to initial gluon
    AP = 1.0/z*(pow(z,4.0)+1.0+pow(1-z,4.0))/z/(1.0-z);
    // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
    AP *= 2.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
  }
  else { // collinear to final quark
    AP = (1.0+pow2(z))/(1.0-z);
    // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
    AP *= 1.0;
    // ant with quark reproduces full AP
    AP *= 1.0;
  }
        
  return AP/Q2;
}

//==========================================================================

// GGEmitIF
// Note: dimensionful parameters [s] = GeV^2

double GGEmitIF::antennaFunction(double saj, double sjk, double sAK, 
                                 int hA, int hK, int ha, int hj, int hk) {

  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hK != 1 && hK != -1 ) hK = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hK == 9) nAvg *= 2;
    
 // Shorthands
  double s         = sAK + sjk;
  double yaj       = saj/s;
  double yjk       = sjk/s;
  double yAK       = sAK/s;
  double eikJ      = 1./(sAK*yaj*yjk);
  double eikA      = 1./(sAK*yaj*(yAK+yaj));
  double colK      = (1.-alphaSav) * (1.-2.*yaj) / (sAK * yjk);  
  
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hK > 0 || hA == 9 || hK == 9) {
    // ++ > +++ && -- > --- antennae (MHV)
    term = eikJ + eikA + colK;
    if ( RH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae (NMHV)
    term = ( pow3(1.-yaj) + pow3(1.-yjk)-1. ) * eikJ
      + ( 6. - 3.*(yaj+yjk) + yaj*yjk )/sAK;
    if ( RH[hA] && RH[hK] && RH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && LH[ha] && RH[hj] && LH[hk] ) hSum += term;
      
    // ++ > --+ && -- > ++- antennae
    term = pow3(yjk) * eikA;
    if ( RH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hK < 0 || hA == 9 || hK == 9) {
    // +- > ++- && -+ > --+ antennae
    term = pow3(1.-yaj) * eikJ + eikA;
    if ( RH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
      
    // +- > +-- && -+ > -++ antennae
    term = pow3(1.-yjk) * eikJ + colK + ( 3.*yaj - yjk - yaj*yjk )/sAK;
    if ( RH[hA] && LH[hK] && RH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && RH[hj] && RH[hk] ) hSum += term;
      
    // +- > --- && -+ > +++ antennae
    term = pow3(yjk) * eikA;
    if ( RH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double GGEmitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
  // Using smaller invariant for collinear limit
  double z  = ( saj < sjk ? z1(saj,sjk,sAK) : z2(saj,sjk,sAK) );
  double Q2 = min(saj,sjk);
  double AP = 0.0;
    
  if (saj < sjk) { // collinear to initial gluon
    AP = 1.0/z*(pow(z,4.0)+1.0+pow(1-z,4.0))/z/(1.0-z);
    // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
    AP *= 2.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
  }
  else { // collinear to final gluon
    AP = (2.0*z/(1.0-z)+z*(1.0-z));
    // AP uses alpha/2pi * CA, ant uses alpha/4pi * CA, with CA = 3
    AP *= 2.0;
    // gluon appears in 2 ants, hence ant reproduces half of AP
    AP *= 0.5;
  }
        
  return AP/Q2;
}

//==========================================================================

// QXSplitIF
// Note: dimensionful parameters [s] = GeV^2
// Note: Splitting in the forwards sense aka quark backwards evolving
//       to a gluon and emitting a antiquark in the final state

double QXSplitIF::antennaFunction(double saj, double sjk, double sAK, 
                                  int hA, int hK, int ha, int hj, int hk) {

  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hK != 1 && hK != -1 ) hK = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hK == 9) nAvg *= 2;
    
  // Shorthands
  double s    = sAK + sjk;
  double yAK  = sAK/s;
  double yaj  = saj/s;
  double colA = 1./(sAK * yaj);
  
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hK > 0 || hA == 9 || hK == 9) {
    // ++ > +-+ && -- > -+- antennae
    term = pow2(yAK) * colA;
    if ( RH[hA] && RH[hK] && RH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && LH[ha] && RH[hj] && LH[hk] ) hSum += term;
      
    // ++ > --+ && -- > ++- antennae
    term = pow2(1.-yAK) * colA;
    if ( RH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hK < 0 || hA == 9 || hK == 9) {
    // +- > +-- && -+ > -++ antennae
    term = pow2(yAK) * colA;
    if ( RH[hA] && LH[hK] && RH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && RH[hj] && RH[hk] ) hSum += term;
      
    // +- > --- && -+ > +++ antennae
    term = pow2(1. - yAK) * colA;
    if ( RH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double QXSplitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                  int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
  // There is only the saj collinear limit
  double z  = z1(saj,sjk,sAK);
  double Q2 = saj;
  double AP = 1.0/z * (pow2(z)+pow2(1-z));
  // AP uses alpha/2pi * TR/2, ant uses alpha/4pi * TR, with TR = 1
  AP *= 1.0;
  // ant with quark reproduces full AP
  AP *= 1.0;
        
  return AP/Q2;
}

//==========================================================================

// GXConvIF
// Note: dimensionful parameters [s] = GeV^2
// Note: Gluon evolves backwards into an IS quark and emits an FS quark

double GXConvIF::antennaFunction(double saj, double sjk, double sAK, 
                                 int hA, int hK, int ha, int hj, int hk) {

  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hK != 1 && hK != -1 ) hK = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hK == 9) nAvg *= 2;
    
  // Shorthands
  double s    = sAK + sjk;
  double yaj  = saj / s;
  double yAK  = sAK / s;
  double colA = 1./(2.*sAK*yaj*(yAK+yaj));
    
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hK > 0 || hA == 9 || hK == 9) {
    // ++ > +++ && -- > --- antennae
    term = colA;
    if ( RH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;
      
    // ++ > --+ && -- > ++- antennae
    term = pow2(1.-yAK) * colA;
    if ( RH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hK < 0 || hA == 9 || hK == 9) {
    // +- > --- && -+ > +++ antennae
    term = pow2(1.-yAK) * colA;
    if ( RH[hA] && LH[hK] && LH[ha] && LH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && RH[ha] && RH[hj] && RH[hk] ) hSum += term;
      
    // +- > ++- && -+ > --+ antennae
    term = colA;
    if ( RH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double GXConvIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                 int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
    
  // There is only the saj collinear limit
  double z  = z1(saj,sjk,sAK);
  double Q2 = saj;
  double AP = 1.0/z * (1.0+pow2(1.0-z))/z;
  // AP uses alpha/2pi * CF, ant uses alpha/4pi * 2CF, with CF = 4/3
  AP *= 1.0;
  // gluon appears in 2 ants, hence ant reproduces half of AP
  AP *= 0.5;
        
  return AP/Q2;
}

//==========================================================================

// XGSplitIF
// Note: dimensionful parameters [s] = GeV^2
// Note: Gluon splitting in the final state

double XGSplitIF::antennaFunction(double saj, double sjk, double sAK, 
                                  int hA, int hK, int ha, int hj, int hk) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;

  // Massless antennae: treat unknown polarisations as unpolarised
  if ( hA != 1 && hA != -1 ) hA = 9;
  if ( hK != 1 && hK != -1 ) hK = 9;
  if ( ha != 1 && ha != -1 ) ha = 9;
  if ( hj != 1 && hj != -1 ) hj = 9;
  if ( hk != 1 && hk != -1 ) hk = 9;
    
  // How many configurations are we averaging over?
  int nAvg = 1;
  if (hA == 9) nAvg *= 2;
  if (hK == 9) nAvg *= 2;
    
  // Shorthands
  double s    = sAK + sjk;
  double yaj  = saj / s;
  double yak  = 1. - yaj;
  double colK = 1./(2*sjk);
  
  // Do helicity sum
  double hSum = 0.0;

  // (++ and --) parents
  if ( hA * hK > 0 || hA == 9 || hK == 9) {
    // ++ > ++- && -- > --+ antennae
    term = pow2(yaj) * colK;
    if ( RH[hA] && RH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && LH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
      
    // ++ > +-+ && -- > -+- antennae
    term = pow2(yak) * colK;
    if ( RH[hA] && RH[hK] && RH[ha] && LH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && LH[hK] && LH[ha] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // (+- and -+) parents
  if ( hA * hK < 0 || hA == 9 || hK == 9) {
    // +- > ++- && -+ > --+ antennae
    term = pow2(yak) * colK;
    if ( RH[hA] && LH[hK] && RH[ha] && RH[hj] && LH[hk] ) hSum += term;
    if ( LH[hA] && RH[hK] && LH[ha] && LH[hj] && RH[hk] ) hSum += term;
      
    // +- > +-+ && -+ > -+- antennae
    term = pow2(yaj) * colK;
    if ( RH[hA] && LH[hK] && RH[ha] && LH[hj] && RH[hk] ) hSum += term; 
    if ( LH[hA] && RH[hK] && LH[ha] && RH[hj] && LH[hk] ) hSum += term;
  }
    
  // Return helicity sum, averaged over initial helicities
  return hSum/nAvg;
}

//--------------------------------------------------------------------------

// The AP kernel: P(z)/Q2

double XGSplitIF::AltarelliParisi(double saj, double sjk, double sAK, 
                                  int, int, int, int, int) {
  // Sanity check. Require positive invariants.
  if ((saj <= 0.0) || (sjk <= 0.0) || (sAK <= 0.0)) return 0.0;
  
  // There is only the sjk collinear limit
  double z  = z2(saj,sjk,sAK);
  double Q2 = sjk;
  double AP = (pow2(z)+pow2(1-z));
  // AP uses alpha/2pi * TR/2, ant uses alpha/4pi * TR, with TR = 1
  AP *= 1.0;
  // gluon appears in 2 ants, hence ant reproduces half of AP
  AP *= 0.5;
        
  return AP/Q2;
}

} // End Vincia Namespace
  
