// Reconnections.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2018 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions for Vincia's Reconnections class.

#include "Vincia/Reconnections.h"

namespace Vincia {

// Initialize
bool Reconnections::init() {

  // Sanity check
  if (!isInitPtr) return false;

  // Set verbose level
  verbose = settingsPtr->mode("Vincia:verbose");

  // Set parameters
  CRmode            = settingsPtr->mode("Vincia:CRmode");
  mMin              = settingsPtr->parm("Vincia:CRmMin");
  m2min             = pow2(mMin); 
  CRjunctions       = settingsPtr->flag("Vincia:CRjunctions");
  junctionsM0       = settingsPtr->parm("Vincia:CRjunctionsM0");
  inheritMode       = settingsPtr->mode("Vincia:CRinheritMode");
  
  // 
  isInit = true;
  return isInit;
}

// Map a list of particles with ordinary Les Houches colour tags
// into a new set of Les Houches colour tags with the last index
// running from 1-9.
bool Reconnections::getColourMap(map<int,int>& colorMap,  
                                       vector<Particle>& particles,
                                       int startTag) {
  // Sanity checks
  if (!isInit) {
    printOut("CR:getColourMap","ERROR! Reconnections not initialized");
    return false;
  } 
  else if (particles.size() <= 1) return false;

  // Construct colour map
  bool doColorize = false;
  int nextTagBase = 10*int((startTag-1)/10);
  for (int i=0; i<int(particles.size()); ++i) {
    int col  = particles[i].col();
    int acol = particles[i].acol();
    if (col == 0 && acol == 0) continue;
    int colIndx = colorMap[col];
    int acolIndx = colorMap[acol];
    nextTagBase += 10;
    if (col != 0 && colIndx == 0) {
      colIndx = nextTagBase + int(rndmPtr->flat()*9) + 1;
      colorMap[col]  = colIndx;
      doColorize = true;
    }
    if (acol != 0 && colIndx == 0) {
      acolIndx = nextTagBase + int(colIndx%10 + rndmPtr->flat()*8)%9 + 1;
      colorMap[acol] = acolIndx;                 
      doColorize = true;
    }
  }
  return doColorize;
}

  // Determine which of two antennae inherits the old colour tag after
// a branching. Default is that the largest invariant has the largest
// probability to inherit, with a few alternatives also implemented.
bool Reconnections::inherit12(double s12, double s23) {
  // Initialization check
  if (!isInit) {
    printOut("VinciaReconnections::inherit12",
             "ERROR! Reconnections not initialized"); 
    if (isInitPtr && rndmPtr->flat() < 0.5) return false;
    else return true;
  }
  // Mode 0: Purely random
  if (inheritMode == 0) {
    if (rndmPtr->flat() < 0.5) return true;
    else return false;
  }
  // Safety checks: small, or approximately equal s12, s23
  double a12 = abs(s12);
  double a23 = abs(s23);
  // Inverted mode (smallest invariant inherits - should only be used for
  // extreme variation checks)
  if (inheritMode < 0) {
    a12 = abs(s23);
    a23 = abs(s12);
    inheritMode = abs(inheritMode);
  }
  // Winner-takes-all mode
  if (inheritMode == 2) {
    if (a12 > a23) return true;
    else return false;
  }
  double p12 = 0.5;
  if ( max(a12,a23) > TINY ) {
    if ( a12 < TINY ) p12 = 0.;
    else if ( a23 < TINY ) p12 = 1.;
    else {
      double r = a23/a12;
      if (r < TINY) p12 = 1. - r;
      else if (r > 1./TINY) p12 = 1./r;
      else p12 = 1./(1. + r);
    }
  }
  if (rndmPtr->flat() < p12) return true;
  else return false;
}

// Dipole swing (perturbative colour reconnections) a la Ariadne. 
int Reconnections::doReconnect(Event& event, 
                                     vector<BranchelementalFSR>& antennae,
                                     double q2now, int measureType) {

  // Sanity checks
  if (!isInit) {
    printOut("CR:doReconnect","ERROR! Reconnections not initialized");
    return 0;
  } 
  else if (antennae.size() <= 1) return 0;

  // Initialize return value: no reconnections done
  int nReconnections = 0;

  // Loop over dipole colour indices.
  for (int iCol=1; iCol <=9; ++iCol) {
    // Create matrix of invariants for this colour index
    vector<int> iAntCol;
    for (int iAnt=0; iAnt<int(antennae.size()); ++iAnt) {
      BranchelementalFSR* antPtr = &antennae[iAnt];
      if (antPtr->old1.col()%10 != iCol) continue;
      iAntCol.push_back(iAnt);
    }
    int nAntennae = iAntCol.size();
    if (nAntennae <= 1) continue;

    // Verbose output
    if (verbose >= 9) {
      event.list();
      cout<<"antenna color index = "<<iCol<<" has "
          <<nAntennae<<" antennae with indices:";
      for (int iAnt=0; iAnt<nAntennae; ++iAnt)
        cout<<" "<<antennae[iAntCol[iAnt]].old1.col();
      cout<<endl;
    }

    // Create list of dot products, with colour-connected invariants on
    // the diagonal
    vector<Vec4> p;
    vector<Vec4> k;
    for (int iAnt = 0; iAnt<nAntennae; ++iAnt) {
      p.push_back(antennae[iAntCol[iAnt]].old1.p());
      k.push_back(antennae[iAntCol[iAnt]].old2.p());
    }
    
    // Create matrix of dot products, check for zeroes (singlet gluons)
    CRmatrix dots;
    bool hasZero = false;
    for (int iAnt=0; iAnt<nAntennae; ++iAnt) {
      if (verbose >= 9) cout<<" "<<iAnt<<" ";        
      for (int jAnt=0; jAnt<nAntennae; ++jAnt) {
        double dot = lambda(p[iAnt],k[jAnt]);
        // Check for zeroes (singlet gluons) and impose minimum m2.
        if (dot <= m2min) hasZero = true;
        if (verbose >= 9) cout<<fixed<<setprecision(3)<<sqrt(abs(2*dot))<<" ";
        // Set entry
        dots.setEntry(iAnt,jAnt,dot);
      }
      if (verbose >= 9) cout<<endl;
    }
    // Do nothing if matrix has a zero
    if (hasZero) {
      if (verbose >= 1) printOut("VS:branch","Failed colour rearrangement");
    } 
    // Look for reconnections
    else {      
      // Find n'th largest invariant on diagonal
      int nAttempts = 0;
      int nReconnectionsNow = 0;
      bool tryReconnect = true;        
      while (tryReconnect && ++nAttempts < 1000) {
        tryReconnect=false;
        for (int iRank = 0; iRank<nAntennae - 1; ++iRank) {
          int i1      = dots.getDiagonalIndexOrdered(iRank);
          if (i1 < 0) break;
          double lamOld1 = dots.getEntry(i1,i1);
          if (verbose >= 9) 
            cout<<" Diagonal index "<<iRank<<" is "<<i1
                <<" with lambda value = "<<lamOld1<<endl;
          // Look for best reconnection in this row
          int i2Min = i1; 
          double measureMin = 1.;
          for (int i2=0; i2<nAntennae; ++i2) {
            if (i2 == i1) continue;
            double lamOld2    = 2. * abs(dots.getEntry(i2,i2));
            double lamProdOld = lamOld1 * lamOld2;
            double lamNew1    = 2. * abs(dots.getEntry(i1,i2));
            double lamNew2    = 2. * abs(dots.getEntry(i2,i1));
            double lamProdNew = lamNew1 * lamNew2;
            bool ok = false;
            double measure = 0;
            // Type 1: Lambda product, with new lambdas > q2now 
            if (measureType == 1) {
              measure = lamProdNew/lamProdOld;
              if (measure < measureMin && min(lamNew1,lamNew2) > q2now) 
                ok = true;
            }
            // Type 2: Lambda product, with new lambdas > q2now and 
            // both lambda values decreasing.
            else if (measureType == 2) {
              measure = lamProdNew/lamProdOld;
              if (measure < measureMin && min(lamNew1,lamNew2) > q2now 
                  && max(lamNew1,lamNew2) < min(lamOld1,lamOld2))
                ok = true;
            } else {
              printOut("Vincia:CR","ERROR! unrecognized measureType");
            }
            // Update smallest-measure reconnection found so far
            if (ok) {
              i2Min      = i2;
              measureMin = measure;
            }
          }
          if (i2Min != i1) {
            if (verbose >= 9) 
              cout<<" Estimated best reconnection is "<<i1<<" <-> "<<i2Min
                  <<" with reduction measure = "<<measureMin<<endl;
            // Do reconnection and restart from beginning
            dots.swapColumns(i1,i2Min);
            nReconnectionsNow++;
            tryReconnect = true;
            break;
          }
        }
      }
      // Rewire the corresponding antennae in the event record and in
      // the list of branchelementals.
      if (nReconnectionsNow >= 1) {
        // Set return value to true: at least one reconnection was done.
        nReconnections += nReconnectionsNow;
        for (int iAnt = 0; iAnt<nAntennae; ++iAnt) {
          int jAnt = dots.getNewColumn(iAnt);
          if (jAnt <= iAnt) continue;
          BranchelementalFSR* antI = &antennae[iAntCol[iAnt]];
          BranchelementalFSR* antJ = &antennae[iAntCol[jAnt]];
          if (verbose >= 8) {
            cout<<" Will reconnect branchelemental "<<iAntCol[iAnt]<<" with "<<iAntCol[jAnt]<<" cols = "<<antI->old1.col()<<" "<<antJ->old1.col()<<endl;
          }
          // I is the first dipole (I1 color, I2 anticolor)
          // J is the second dipole (J1 color, J2 anticolor)
          int colI    = antI->old1.col();
          int colJ    = antJ->old1.col();
          int iOld1   = antI->iOld1;
          int iOld2   = antI->iOld2;
          int jOld1   = antJ->iOld1;
          int jOld2   = antJ->iOld2;
          int iNew1   = iOld1;
          int iNew2   = jOld2;
          int jNew1   = jOld1;
          int jNew2   = iOld2;
          // Also save the old IL and IR neighbours
          int iOldL   = antI->iL;
          int iOldR   = antI->iR;
          int jOldL   = antJ->iL;
          int jOldR   = antJ->iR;
          // Rewire the I and J colours, reset branchelementals
          event[iNew1].col(colI);
          event[iNew2].acol(colI);
          event[jNew1].col(colJ);
          event[jNew2].acol(colJ);
          antI->reset(event,iNew1,iNew2);
          antJ->reset(event,jNew1,jNew2);   
          // Set new I and J neighbours (same Ls, swap Rs)
          antI->iL    = iOldL;
          antI->iR    = jOldR;
          antJ->iL    = jOldL;
          antJ->iR    = iOldR;
          // Quarks cannot have cyclic colour flows
          if (antI->old1.isQuark() && antI->iL == antI->iOld2) antI->iL = 0;
          if (antI->old2.isQuark() && antI->iR == antI->iOld1) antI->iR = 0;
          if (antJ->old1.isQuark() && antJ->iL == antJ->iOld2) antJ->iL = 0;
          if (antJ->old2.isQuark() && antJ->iR == antJ->iOld1) antJ->iR = 0;
          // Loop over all antennae, update other neighbours
          for (int kAnt=0; kAnt<int(antennae.size()); ++kAnt) {
            BranchelementalFSR* kPtr = &antennae[kAnt];
            // Skip the ones we just reconnected: neighbours already set
            if (kPtr == antI || kPtr == antJ) continue;
            int iL  = kPtr->iL;
            int iR  = kPtr->iR;
            // Update L neighbours
            if (iL == iOld1) kPtr->iL  = jOld1;
            else if (iL == jOld1) kPtr->iL  = iOld1;
            // Update R neighbours
            if (iR == iOld2) kPtr->iR  = jOld2;
            else if (iR == jOld2) kPtr->iR  = iOld2;
            // Quarks cannot have cyclic flows
            if (kPtr->old1.isQuark() && kPtr->iL == kPtr->iOld2) kPtr->iL = 0;
            if (kPtr->old2.isQuark() && kPtr->iR == kPtr->iOld1) kPtr->iR = 0;
          }
        }
      }          
    }
  }
  if (nReconnections >= 1 && verbose >= 9) {
    cout<<"New event:"<<endl;
    event.list();
  }
  // Return
  return nReconnections;
}

// Dipole swing (perturbative colour reconnections) a la Ariadne. 
int Reconnections::doJunctions(Event& event, 
				     vector<BranchelementalFSR>& antennae){

  // Sanity checks
  if (!isInit) {
    printOut("CR:doJunctions","ERROR! Reconnections not initialized");
    return 0;
  } 
  else if (antennae.size() <= 1) return 0;

  // Find the number of independent colour chains
  vector<int> iAntChainBegin;
  for (int iAnt=0; iAnt < int(antennae.size()); ++iAnt) {
    if (antennae[iAnt].old1.colType() == 1) iAntChainBegin.push_back(iAnt);    
  }
  int nChains = iAntChainBegin.size();

  // Junction reconnections can only occur if at least two chains are present
  if (nChains <= 1) return 0;

  if (verbose >= 9) {
    event.list();
    printOut("CR:doJunctions","Testing junction reconnections");
  }

  // Go through all chain possibilities. Find best junction reconnection
  // First chain is [1,n-1]
  double ratioMin = 1e99;
  int    iAntMin    = -1;
  int    jAntMin    = -1;
  for (int iChain=0; iChain<nChains-1; ++iChain) {
    // Verbose output
    if (verbose >= 9) printOut("CR:doJunctions","Next antenna chain");
    // Start from first antenna in iChain
    int iAnt  = iAntChainBegin[iChain];
    int iCol  = antennae[iAnt].iOld1;
    int iAcol = antennae[iAnt].iOld2;
    // Now step through the individual antennae in iChain, until end of chain
    bool endOfChainI = false;
    while (!endOfChainI) {
      int iTag  = antennae[iAnt].old1.col();
      Vec4 pI   = event[iCol].p();
      Vec4 kI   = event[iAcol].p();
      double lamII = lambda(pI,kI);
      int iL    = (event[iCol].colType() == 2) ? antennae[iAnt].iL : 0;
      int iR    = (event[iAcol].colType() == 2) ? antennae[iAnt].iR : 0;
      // Check for small neighbour invariants 
      double lamIL = (iL > 0) ? lambda(pI,event[iL].p()) : 0.0;
      double lamIR = (iR > 0) ? lambda(kI,event[iR].p()) : 0.0;
      if (verbose >= 9) {
	cout<<" I : "<<iCol<<" "<<iAcol<<" tag "<<antennae[iAnt].old1.col()
	    <<" m = "<<sqrt(lamII)<<" mN(L,R) = "<<sqrt(lamIL)<<" "<<sqrt(lamIR)<<endl;
      }

      // Second chain is [iChain+1,n]
      for (int jChain=iChain+1; jChain<nChains; ++jChain) {
	// Start from first antenna in jChain
 	int jAnt  = iAntChainBegin[jChain];
	int jCol  = antennae[jAnt].iOld1;
	int jAcol = antennae[jAnt].iOld2;
	// Now step through the individual antennae in jChain
	bool endOfChainJ = false;
	while (!endOfChainJ) {
	  int jTag  = antennae[jAnt].old1.col();
	  // Only allow reconnections between opposite-parity antenna indices
	  if (iTag%2 != jTag%2) {
	    Vec4 pJ   = event[jCol].p();
	    Vec4 kJ   = event[jAcol].p();
	    double lamJJ = lambda(pJ,kJ);
	    int jL    = (event[jCol].colType() == 2) ? antennae[jAnt].iL : 0;
	    int jR    = (event[jAcol].colType() == 2) ? antennae[jAnt].iR : 0;
	    // Check for small neighbour invariants 
	    double lamJL = (jL > 0) ? lambda(pJ,event[jL].p()) : 0.0;
	    double lamJR = (jR > 0) ? lambda(kJ,event[jR].p()) : 0.0;
	    if (verbose >= 9) {
	      cout<<" J : "<<jCol<<" "<<jAcol<<" tag "<<antennae[jAnt].old1.col()
		  <<" m = "<<sqrt(lamJJ)<<" mN(L,R) = "<<sqrt(lamJL)<<" "<<sqrt(lamJR)<<endl;
	    }

	    double orgMeasure = lamII * lamJJ;
	    double newMeasure = 8 * lambda(pI+pJ,kI+kJ) 
	      * lambda(pI,pJ) * lambda(kI,kJ) / pow2(junctionsM0);
	    double ratio = newMeasure/orgMeasure;
	    if (ratio < ratioMin) {
	      ratioMin = ratio;
	      iAntMin = iAnt;
	      jAntMin = jAnt;
	    }
	  }
	  
	  // End of iChain: antiquark
	  if (antennae[jAnt].old2.colType() == -1) {
	    endOfChainJ = true;
	  }
	  // Else step across gluon in jChain: update jCol, jAcol, and jAnt
	  else if (antennae[jAnt].old2.colType() == 2) {
	    jCol  = antennae[jAnt].iOld2;
	    jAcol = antennae[jAnt].iR;
	    // Update jAnt: find the antenna that has these parents
	    bool foundNeighbourJ = false;
	    for (int kAnt=0; kAnt<int(antennae.size()); ++kAnt) {
	      if (kAnt == jAnt) continue;
	      if (antennae[kAnt].iOld1 == jCol 
		  && antennae[kAnt].iOld2 == jAcol) {
		jAnt = kAnt;
		foundNeighbourJ = true;
	      }
	    }
	    if (!foundNeighbourJ) {
	      printOut("CR:doReconnectJunction",
		       "failed to find colour neighbour");
	      return 0;
	    }
	  }
	  else {
	    printOut("CR:doReconnectJunction","failed to follow colour chain");
	    return 0;
	  }

	} // End of iteration through the individual antennae in jChain
	  
      } // End loop over index of second chain: jChain
      
      // End of iChain: antiquark
      if (antennae[iAnt].old2.colType() == -1) {
	endOfChainI = true;
      }
      // Else step across gluon in iChain: update iCol, iAcol, and iAnt
      else if (antennae[iAnt].old2.colType() == 2) {
	iCol  = antennae[iAnt].iOld2;
	iAcol = antennae[iAnt].iR;
	// Update iAnt: find the antenna that has these parents
	bool foundNeighbourI = false;
	for (int kAnt=0; kAnt<int(antennae.size()); ++kAnt) {
	  if (kAnt == iAnt) continue;
	  if (antennae[kAnt].iOld1 == iCol && antennae[kAnt].iOld2 == iAcol) {
	    iAnt = kAnt;
	    foundNeighbourI = true;
	  }
	}
	if (!foundNeighbourI) {
	  printOut("CR:doReconnectJunction","failed to find colour neighbour");
	  return 0;
	}
      }
      else {
	printOut("CR:doReconnectJunction","failed to follow colour chain");
	return 0;
      }
    } // End of iteration through the individual antennae in iChain

  } // End loop over index of first chain: iChain

  // Exit if no good junction configurations found
  if (ratioMin > 1.) return 0;

  // Do junction reconnection (so far can at most do one, even if there
  // are multiple chains)
  int iCol     = antennae[iAntMin].iOld1;
  int iAcol    = antennae[iAntMin].iOld2;
  int iTagCol  = antennae[iAntMin].old1.col();
  int jCol     = antennae[jAntMin].iOld1;
  int jAcol    = antennae[jAntMin].iOld2;
  int jTagCol  = antennae[jAntMin].old1.col();
  int kTag     = event.nextColTag();
  int iTagAcol = event.nextColTag();
  int jTagAcol = event.nextColTag();
  event[iAcol].acol(iTagAcol);
  event[jAcol].acol(jTagAcol);
  event.appendJunction( 1, iTagCol, jTagCol, kTag);
  event.appendJunction( 2, iTagAcol, jTagAcol, kTag);
  
  // Erase I and J antennae from list of branchelementals  
  antennae.erase(antennae.begin()+max(iAntMin,jAntMin));
  antennae.erase(antennae.begin()+min(iAntMin,jAntMin));

  // Verbose output
  if (verbose >= 9) {
    event.list();
    cout<<"Rewired antennae I "<<iCol<<" "<<iAcol<<" and J "<<jCol<<" "<<jAcol
	<<" as new JJ system with tags "<<iTagCol<<" "<<jTagCol<<" "<<kTag
	<<" "<<iTagAcol<<" "<<jTagAcol<<endl;
  }

  // Return
  return 0;
}

// Print header information from colour reconnections
void Reconnections::header(int len) {
  cout << " | Colour\n"
       << " | Reconnections:  CRmode               = "<<
    num2str(CRmode,len)<<"  (0:off, 1:non-perturbative only, 2:all)\n";
  if (CRmode >= 1) {
    cout << " |                 CRmMin               = "
	 << num2str(mMin,len)<<"\n"
	 << " |                 CRjunctions          = "
	 << bool2str(CRjunctions,len)<<"\n";
    if (CRjunctions) {
      cout << " |                 CRjunctionsM0        ="
	   << num2str(junctionsM0,len)<<"\n";
    }
  }
  
}

// End namespace Vincia
}

